#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import copy
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
)
from uuid import UUID

import sqlalchemy
from fastapi.encoders import jsonable_encoder
from sqlalchemy import Select, and_, func, select
from sqlalchemy.orm import contains_eager, joinedload, selectinload

import argilla_server.errors.future as errors
from argilla_server.contexts import accounts
from argilla_server.enums import DatasetStatus, RecordInclude, UserRole
from argilla_server.models import (
    Dataset,
    Field,
    MetadataProperty,
    Question,
    Record,
    Response,
    ResponseStatus,
    Suggestion,
    Vector,
    VectorSettings,
)
from argilla_server.models.suggestions import SuggestionCreateWithRecordId
from argilla_server.schemas.v0.users import User
from argilla_server.schemas.v1.datasets import (
    DatasetCreate,
)
from argilla_server.schemas.v1.fields import FieldCreate
from argilla_server.schemas.v1.metadata_properties import MetadataPropertyCreate, MetadataPropertyUpdate
from argilla_server.schemas.v1.questions import QuestionCreate
from argilla_server.schemas.v1.records import (
    RecordCreate,
    RecordIncludeParam,
    RecordsCreate,
    RecordsUpdate,
    RecordUpdateWithId,
)
from argilla_server.schemas.v1.responses import (
    ResponseCreate,
    ResponseUpdate,
    ResponseUpsert,
    ResponseValueCreate,
    ResponseValueUpdate,
)
from argilla_server.schemas.v1.vector_settings import (
    VectorSettings as VectorSettingsSchema,
)
from argilla_server.schemas.v1.vector_settings import (
    VectorSettingsCreate,
)
from argilla_server.schemas.v1.vectors import Vector as VectorSchema
from argilla_server.search_engine import SearchEngine

if TYPE_CHECKING:
    from sqlalchemy.ext.asyncio import AsyncSession

    from argilla_server.schemas.v1.datasets import (
        DatasetUpdate,
    )
    from argilla_server.schemas.v1.fields import FieldUpdate
    from argilla_server.schemas.v1.questions import QuestionUpdate
    from argilla_server.schemas.v1.records import RecordUpdate
    from argilla_server.schemas.v1.suggestions import SuggestionCreate
    from argilla_server.schemas.v1.vector_settings import VectorSettingsUpdate

LIST_RECORDS_LIMIT = 20

VISIBLE_FOR_ANNOTATORS_ALLOWED_ROLES = [UserRole.admin, UserRole.annotator]
NOT_VISIBLE_FOR_ANNOTATORS_ALLOWED_ROLES = [UserRole.admin]


async def _touch_dataset_last_activity_at(db: "AsyncSession", dataset: Dataset) -> Dataset:
    return await db.execute(
        sqlalchemy.update(Dataset).where(Dataset.id == dataset.id).values(last_activity_at=datetime.utcnow())
    )


async def get_dataset_by_id(
    db: "AsyncSession",
    dataset_id: UUID,
    with_fields: bool = False,
    with_questions: bool = False,
    with_metadata_properties: bool = False,
    with_vectors_settings: bool = False,
) -> Dataset:
    query = select(Dataset).filter_by(id=dataset_id)
    options = []
    if with_fields:
        options.append(selectinload(Dataset.fields))
    if with_questions:
        options.append(selectinload(Dataset.questions))
    if with_metadata_properties:
        options.append(selectinload(Dataset.metadata_properties))
    if with_vectors_settings:
        options.append(selectinload(Dataset.vectors_settings))
    if options:
        query = query.options(*options)
    result = await db.execute(query)
    return result.scalar_one_or_none()


async def get_dataset_by_name_and_workspace_id(
    db: "AsyncSession", name: str, workspace_id: UUID
) -> Union[Dataset, None]:
    result = await db.execute(select(Dataset).filter_by(name=name, workspace_id=workspace_id))
    return result.scalar_one_or_none()


async def list_datasets(db: "AsyncSession") -> List[Dataset]:
    result = await db.execute(select(Dataset).order_by(Dataset.inserted_at.asc()))
    return result.scalars().all()


async def list_datasets_by_workspace_id(db: "AsyncSession", workspace_id: UUID) -> List[Dataset]:
    result = await db.execute(
        select(Dataset).where(Dataset.workspace_id == workspace_id).order_by(Dataset.inserted_at.asc())
    )
    return result.scalars().all()


async def create_dataset(db: "AsyncSession", dataset_create: DatasetCreate):
    return await Dataset.create(
        db,
        name=dataset_create.name,
        guidelines=dataset_create.guidelines,
        allow_extra_metadata=dataset_create.allow_extra_metadata,
        workspace_id=dataset_create.workspace_id,
    )


async def _count_required_fields_by_dataset_id(db: "AsyncSession", dataset_id: UUID) -> int:
    result = await db.execute(select(func.count(Field.id)).filter_by(dataset_id=dataset_id, required=True))
    return result.scalar()


async def _count_required_questions_by_dataset_id(db: "AsyncSession", dataset_id: UUID) -> int:
    result = await db.execute(select(func.count(Question.id)).filter_by(dataset_id=dataset_id, required=True))
    return result.scalar()


def _allowed_roles_for_metadata_property_create(metadata_property_create: MetadataPropertyCreate) -> List[UserRole]:
    if metadata_property_create.visible_for_annotators:
        return VISIBLE_FOR_ANNOTATORS_ALLOWED_ROLES
    else:
        return NOT_VISIBLE_FOR_ANNOTATORS_ALLOWED_ROLES


async def publish_dataset(db: "AsyncSession", search_engine: SearchEngine, dataset: Dataset) -> Dataset:
    if dataset.is_ready:
        raise ValueError("Dataset is already published")

    if await _count_required_fields_by_dataset_id(db, dataset.id) == 0:
        raise ValueError("Dataset cannot be published without required fields")

    if await _count_required_questions_by_dataset_id(db, dataset.id) == 0:
        raise ValueError("Dataset cannot be published without required questions")

    async with db.begin_nested():
        dataset = await dataset.update(db, status=DatasetStatus.ready, autocommit=False)
        await search_engine.create_index(dataset)

    await db.commit()

    return dataset


async def delete_dataset(db: "AsyncSession", search_engine: SearchEngine, dataset: Dataset) -> Dataset:
    async with db.begin_nested():
        dataset = await dataset.delete(db, autocommit=False)
        await search_engine.delete_index(dataset)

    await db.commit()

    return dataset


async def update_dataset(db: "AsyncSession", dataset: Dataset, dataset_update: "DatasetUpdate") -> Dataset:
    params = dataset_update.dict(exclude_unset=True)
    return await dataset.update(db, **params)


async def get_field_by_id(db: "AsyncSession", field_id: UUID) -> Union[Field, None]:
    result = await db.execute(select(Field).filter_by(id=field_id).options(selectinload(Field.dataset)))
    return result.scalar_one_or_none()


async def get_field_by_name_and_dataset_id(db: "AsyncSession", name: str, dataset_id: UUID) -> Union[Field, None]:
    result = await db.execute(select(Field).filter_by(name=name, dataset_id=dataset_id))
    return result.scalar_one_or_none()


async def create_field(db: "AsyncSession", dataset: Dataset, field_create: FieldCreate) -> Field:
    if dataset.is_ready:
        raise ValueError("Field cannot be created for a published dataset")

    return await Field.create(
        db,
        name=field_create.name,
        title=field_create.title,
        required=field_create.required,
        settings=field_create.settings.dict(),
        dataset_id=dataset.id,
    )


async def update_field(db: "AsyncSession", field: Field, field_update: "FieldUpdate") -> Field:
    params = field_update.dict(exclude_unset=True)
    return await field.update(db, **params)


async def delete_field(db: "AsyncSession", field: Field) -> Field:
    if field.dataset.is_ready:
        raise ValueError("Fields cannot be deleted for a published dataset")

    return await field.delete(db)


async def get_question_by_id(db: "AsyncSession", question_id: UUID) -> Union[Question, None]:
    result = await db.execute(select(Question).filter_by(id=question_id).options(selectinload(Question.dataset)))
    return result.scalar_one_or_none()


async def get_question_by_name_and_dataset_id(db: "AsyncSession", name: str, dataset_id: UUID) -> Union[Question, None]:
    result = await db.execute(select(Question).filter_by(name=name, dataset_id=dataset_id))

    return result.scalar_one_or_none()


async def get_question_by_name_and_dataset_id_or_raise(db: "AsyncSession", name: str, dataset_id: UUID) -> Question:
    question = await get_question_by_name_and_dataset_id(db, name, dataset_id)
    if question is None:
        raise errors.NotFoundError(f"Question with name `{name}` not found for dataset with id `{dataset_id}`")

    return question


async def get_metadata_property_by_name_and_dataset_id(
    db: "AsyncSession", name: str, dataset_id: UUID
) -> Union[MetadataProperty, None]:
    result = await db.execute(select(MetadataProperty).filter_by(name=name, dataset_id=dataset_id))

    return result.scalar_one_or_none()


async def get_metadata_property_by_name_and_dataset_id_or_raise(
    db: "AsyncSession", name: str, dataset_id: UUID
) -> MetadataProperty:
    metadata_property = await get_metadata_property_by_name_and_dataset_id(db, name, dataset_id)
    if metadata_property is None:
        raise errors.NotFoundError(f"Metadata property with name `{name}` not found for dataset with id `{dataset_id}`")

    return metadata_property


async def delete_metadata_property(db: "AsyncSession", metadata_property: MetadataProperty) -> MetadataProperty:
    return await metadata_property.delete(db)


async def create_question(db: "AsyncSession", dataset: Dataset, question_create: QuestionCreate) -> Question:
    if dataset.is_ready:
        raise ValueError("Question cannot be created for a published dataset")

    return await Question.create(
        db,
        name=question_create.name,
        title=question_create.title,
        description=question_create.description,
        required=question_create.required,
        settings=question_create.settings.dict(),
        dataset_id=dataset.id,
    )


async def create_metadata_property(
    db: "AsyncSession",
    search_engine: "SearchEngine",
    dataset: Dataset,
    metadata_property_create: MetadataPropertyCreate,
) -> MetadataProperty:
    async with db.begin_nested():
        metadata_property = await MetadataProperty.create(
            db,
            name=metadata_property_create.name,
            title=metadata_property_create.title,
            settings=metadata_property_create.settings.dict(),
            allowed_roles=_allowed_roles_for_metadata_property_create(metadata_property_create),
            dataset_id=dataset.id,
            autocommit=False,
        )
        if dataset.is_ready:
            await db.flush([metadata_property])
            await search_engine.configure_metadata_property(dataset, metadata_property)

    await db.commit()

    return metadata_property


async def update_metadata_property(
    db: "AsyncSession",
    metadata_property: MetadataProperty,
    metadata_property_update: MetadataPropertyUpdate,
):
    return await metadata_property.update(
        db,
        title=metadata_property_update.title or metadata_property.title,
        allowed_roles=_allowed_roles_for_metadata_property_create(metadata_property_update),
    )


async def update_question(db: "AsyncSession", question: Question, question_update: "QuestionUpdate") -> Question:
    params = question_update.dict(exclude_unset=True)
    return await question.update(db, **params)


async def delete_question(db: "AsyncSession", question: Question) -> Question:
    if question.dataset.is_ready:
        raise ValueError("Questions cannot be deleted for a published dataset")

    return await question.delete(db)


async def count_vectors_settings_by_dataset_id(db: "AsyncSession", dataset_id: UUID) -> int:
    result = await db.execute(select(func.count(VectorSettings.id)).filter_by(dataset_id=dataset_id))

    return result.scalar()


async def get_vector_settings_by_id(db: "AsyncSession", vector_settings_id: UUID) -> Union[VectorSettings, None]:
    result = await db.execute(
        select(VectorSettings).filter_by(id=vector_settings_id).options(selectinload(VectorSettings.dataset))
    )
    return result.scalar_one_or_none()


async def get_vector_settings_by_name_and_dataset_id(
    db: "AsyncSession", name: str, dataset_id: UUID
) -> Union[VectorSettings, None]:
    return await VectorSettings.read_by(db, name=name, dataset_id=dataset_id)


async def update_vector_settings(
    db: "AsyncSession", vector_settings: VectorSettings, vector_settings_update: "VectorSettingsUpdate"
) -> VectorSettings:
    params = vector_settings_update.dict(exclude_unset=True)
    return await vector_settings.update(db, **params)


async def delete_vector_settings(db: "AsyncSession", vector_settings: VectorSettings) -> VectorSettings:
    # TODO: for now the search engine does not allow to delete vector settings
    return await vector_settings.delete(db)


async def create_vector_settings(
    db: "AsyncSession", search_engine: "SearchEngine", dataset: Dataset, vector_settings_create: "VectorSettingsCreate"
) -> VectorSettings:
    async with db.begin_nested():
        vector_settings = await VectorSettings.create(
            db,
            name=vector_settings_create.name,
            title=vector_settings_create.title,
            dimensions=vector_settings_create.dimensions,
            dataset_id=dataset.id,
            autocommit=False,
        )

        if dataset.is_ready:
            await db.flush([vector_settings])
            await search_engine.configure_index_vectors(vector_settings)

    await db.commit()

    return vector_settings


async def get_record_by_id(
    db: "AsyncSession",
    record_id: UUID,
    with_dataset: bool = False,
    with_suggestions: bool = False,
    with_vectors: bool = False,
) -> Union[Record, None]:
    query = select(Record).filter_by(id=record_id)
    if with_dataset:
        query = query.options(
            selectinload(Record.dataset).selectinload(Dataset.questions),
            selectinload(Record.dataset).selectinload(Dataset.metadata_properties),
        )

    if with_suggestions:
        query = query.options(selectinload(Record.suggestions))
    if with_vectors:
        query = query.options(selectinload(Record.vectors))
    result = await db.execute(query)

    return result.scalar_one_or_none()


async def get_records_by_ids(
    db: "AsyncSession",
    records_ids: Iterable[UUID],
    dataset_id: Optional[UUID] = None,
    include: Optional["RecordIncludeParam"] = None,
    user_id: Optional[UUID] = None,
) -> List[Union[Record, None]]:
    query = select(Record)

    if dataset_id:
        query.filter(Record.dataset_id == dataset_id)

    query = query.filter(Record.id.in_(records_ids))

    if include and include.with_responses:
        if not user_id:
            query = query.options(joinedload(Record.responses))
        else:
            query = query.outerjoin(
                Response, and_(Response.record_id == Record.id, Response.user_id == user_id)
            ).options(contains_eager(Record.responses))

    query = await _configure_query_relationships(query=query, dataset_id=dataset_id, include_params=include)

    result = await db.execute(query)
    records = result.unique().scalars().all()

    # Preserve the order of the `record_ids` list
    record_order_map = {record.id: record for record in records}
    ordered_records = [record_order_map.get(record_id, None) for record_id in records_ids]

    return ordered_records


async def _configure_query_relationships(
    query: "Select", dataset_id: UUID, include_params: Optional["RecordIncludeParam"] = None
) -> "Select":
    if not include_params:
        return query

    if include_params.with_suggestions:
        query = query.options(joinedload(Record.suggestions))

    if include_params.with_all_vectors:
        query = query.options(joinedload(Record.vectors).joinedload(Vector.vector_settings))

    elif include_params.with_some_vector:
        vector_settings_ids_subquery = select(VectorSettings.id).filter(
            and_(VectorSettings.dataset_id == dataset_id, VectorSettings.name.in_(include_params.vectors))
        )
        query = query.outerjoin(
            Vector, and_(Vector.record_id == Record.id, Vector.vector_settings_id.in_(vector_settings_ids_subquery))
        ).options(contains_eager(Record.vectors).joinedload(Vector.vector_settings))

    return query


async def count_records_by_dataset_id(db: "AsyncSession", dataset_id: UUID) -> int:
    result = await db.execute(select(func.count(Record.id)).filter_by(dataset_id=dataset_id))
    return result.scalar()


_EXTRA_METADATA_FLAG = "extra"


async def _validate_metadata(
    db: "AsyncSession",
    dataset: Dataset,
    metadata: Dict[str, Any],
    metadata_properties: Optional[Dict[str, Union[MetadataProperty, Literal["extra"]]]] = None,
) -> Dict[str, Union[MetadataProperty, Literal["extra"]]]:
    if metadata_properties is None:
        metadata_properties = {}

    for name, value in metadata.items():
        metadata_property = metadata_properties.get(name)

        if metadata_property is None:
            metadata_property = await get_metadata_property_by_name_and_dataset_id(db, name=name, dataset_id=dataset.id)

            # If metadata property does not exists but extra metadata is allowed, then we set a flag value to
            # avoid querying the database again
            if metadata_property is None and dataset.allow_extra_metadata:
                metadata_property = _EXTRA_METADATA_FLAG
                metadata_properties[name] = metadata_property
            elif metadata_property is not None:
                metadata_properties[name] = metadata_property
            else:
                raise ValueError(
                    f"'{name}' metadata property does not exists for dataset '{dataset.id}' and extra metadata is"
                    " not allowed for this dataset"
                )

        # If metadata property is not found and extra metadata is allowed, then we skip the value validation
        if metadata_property == _EXTRA_METADATA_FLAG:
            continue

        try:
            if value is not None:
                metadata_property.parsed_settings.check_metadata(value)
        except ValueError as e:
            raise ValueError(f"'{name}' metadata property validation failed because {e}") from e

    return metadata_properties


async def _validate_suggestion(
    db: "AsyncSession",
    suggestion: "SuggestionCreate",
    questions: Optional[Dict[UUID, Question]] = None,
) -> Dict[UUID, Question]:
    if not questions:
        questions = {}

    question = questions.get(suggestion.question_id, None)

    if not question:
        question = await get_question_by_id(db, suggestion.question_id)
        if not question:
            raise ValueError(f"question_id={str(suggestion.question_id)} does not exist")
        questions[suggestion.question_id] = question

    question.parsed_settings.check_response(suggestion)

    return questions


async def validate_user(db: "AsyncSession", user_id: UUID, users_ids: Optional[Set[UUID]]) -> Set[UUID]:
    if not users_ids:
        users_ids = set()

    if user_id not in users_ids:
        if not await accounts.user_exists(db, user_id):
            raise ValueError(f"user_id={str(user_id)} does not exist")
        users_ids.add(user_id)

    return users_ids


async def _validate_vector(
    db: "AsyncSession",
    dataset_id: UUID,
    vector_name: str,
    vector_value: List[float],
    vectors_settings: Optional[Dict[str, VectorSettingsSchema]] = None,
) -> Dict[str, VectorSettingsSchema]:
    if vectors_settings is None:
        vectors_settings = {}

    vector_settings = vectors_settings.get(vector_name, None)
    if not vector_settings:
        vector_settings = await get_vector_settings_by_name_and_dataset_id(db, vector_name, dataset_id)
        if not vector_settings:
            raise ValueError(f"vector with name={str(vector_name)} does not exist for dataset_id={str(dataset_id)}")

        vector_settings = VectorSettingsSchema.from_orm(vector_settings)
        vectors_settings[vector_name] = vector_settings

    vector_settings.check_vector(vector_value)

    return vectors_settings


async def _create_record(
    db: "AsyncSession", dataset: Dataset, record_create: RecordCreate, caches: Dict[str, Any]
) -> Record:
    _validate_record_fields(dataset, fields=record_create.fields)
    caches["metadata_properties_cache"] = await _validate_record_metadata(
        db, dataset, record_create.metadata, caches["metadata_properties_cache"]
    )
    record_responses, caches["users_ids_cache"] = await _build_record_responses(
        db, dataset, record_create, caches["users_ids_cache"]
    )
    record_suggestions, caches["questions_cache"] = await _build_record_suggestions(
        db, record_create, caches["questions_cache"]
    )
    record_vectors, caches["vectors_settings_cache"] = await _build_record_vectors(
        db,
        dataset,
        record_create,
        build_vector_func=lambda value, vector_settings_id: Vector(value=value, vector_settings_id=vector_settings_id),
        cache=caches["vectors_settings_cache"],
    )
    record = Record(
        fields=record_create.fields,
        metadata_=record_create.metadata,
        external_id=record_create.external_id,
        dataset_id=dataset.id,
        responses=record_responses,
        suggestions=record_suggestions,
        vectors=record_vectors,
    )
    return record


async def create_records(
    db: "AsyncSession", search_engine: SearchEngine, dataset: Dataset, records_create: RecordsCreate
):
    if not dataset.is_ready:
        raise ValueError("Records cannot be created for a non published dataset")

    records = []

    caches = {
        "users_ids_cache": set(),
        "questions_cache": {},
        "metadata_properties_cache": {},
        "vectors_settings_cache": {},
    }

    for record_i, record_create in enumerate(records_create.items):
        try:
            record = await _create_record(db, dataset, record_create, caches)
        except ValueError as e:
            raise ValueError(f"Record at position {record_i} is not valid because {e}") from e
        records.append(record)

    async with db.begin_nested():
        db.add_all(records)
        await db.flush(records)
        await _preload_records_relationships_before_index(db, records)
        await search_engine.index_records(dataset, records)

    await db.commit()


async def _load_users_from_responses(responses: Union[Response, Iterable[Response]]) -> None:
    if isinstance(responses, Response):
        responses = [responses]

    # TODO: We should do a single query retrieving all the users from all responses instead of using awaitable_attrs,
    # something similar to what we are already doing in _preload_suggestion_relationships_before_index.
    for response in responses:
        await response.awaitable_attrs.user


async def _load_users_from_record_responses(records: Iterable[Record]) -> None:
    for record in records:
        await _load_users_from_responses(record.responses)


async def _validate_record_metadata(
    db: "AsyncSession",
    dataset: Dataset,
    metadata: Optional[Dict[str, Any]] = None,
    cache: Dict[str, Union[MetadataProperty, Literal["extra"]]] = {},
) -> Dict[str, Union[MetadataProperty, Literal["extra"]]]:
    """Validate metadata for a record."""
    if not metadata:
        return cache

    try:
        cache = await _validate_metadata(db, dataset=dataset, metadata=metadata, metadata_properties=cache)
        return cache
    except ValueError as e:
        raise ValueError(f"metadata is not valid: {e}") from e


async def _build_record_responses(
    db: "AsyncSession",
    dataset: Dataset,
    record_create: RecordCreate,
    cache: Set[UUID] = set(),
) -> Tuple[List[Response], Set[UUID]]:
    """Create responses for a record."""
    if not record_create.responses:
        return [], cache

    responses = []

    for idx, response in enumerate(record_create.responses):
        try:
            cache = await validate_user(db, response.user_id, cache)
            _validate_response_values(dataset, values=response.values, status=response.status)
            responses.append(
                Response(
                    values=jsonable_encoder(response.values),
                    status=response.status,
                    user_id=response.user_id,
                )
            )
        except ValueError as e:
            raise ValueError(f"response at position {idx} is not valid: {e}") from e

    return responses, cache


async def _build_record_suggestions(
    db: "AsyncSession",
    record_schema: Union["RecordCreate", "RecordUpdate", "RecordUpdateWithId"],
    cache: Dict[UUID, Question] = {},
) -> Tuple[List[Suggestion], Dict[UUID, Question]]:
    """Create suggestions for a record."""

    if not record_schema.suggestions:
        return [], cache

    suggestions = []
    for suggestion in record_schema.suggestions:
        try:
            cache = await _validate_suggestion(db, suggestion, questions=cache)
            suggestion = Suggestion(
                type=suggestion.type,
                score=suggestion.score,
                value=suggestion.value,
                agent=suggestion.agent,
                question_id=suggestion.question_id,
            )
            if isinstance(record_schema, RecordUpdateWithId):
                suggestion.record_id = record_schema.id
            suggestions.append(suggestion)
        except ValueError as e:
            raise ValueError(f"suggestion for question_id={suggestion.question_id} is not valid: {e}") from e
    return suggestions, cache


VectorClass = TypeVar("VectorClass")


async def _build_record_vectors(
    db: "AsyncSession",
    dataset: Dataset,
    record_schema: Union["RecordCreate", "RecordUpdate", "RecordUpdateWithId"],
    build_vector_func: Callable[[List[float], UUID], VectorClass],
    cache: Dict[str, VectorSettingsSchema] = {},
) -> Tuple[List[VectorClass], Dict[str, VectorSettingsSchema]]:
    """Create vectors for a record."""
    if not record_schema.vectors:
        return [], cache

    vectors = []
    for vector_name, vector_value in record_schema.vectors.items():
        try:
            cache = await _validate_vector(db, dataset.id, vector_name, vector_value, vectors_settings=cache)
            vectors.append(build_vector_func(vector_value, cache[vector_name].id))
        except ValueError as e:
            raise ValueError(f"vector with name={vector_name} is not valid: {e}") from e
    return vectors, cache


async def _exists_records_with_ids(db: "AsyncSession", dataset_id: UUID, records_ids: List[UUID]) -> List[UUID]:
    result = await db.execute(select(Record.id).filter(Record.dataset_id == dataset_id, Record.id.in_(records_ids)))
    return result.scalars().all()


async def _update_record(
    db: "AsyncSession", dataset: Dataset, record_update: "RecordUpdateWithId", caches: Optional[Dict[str, Any]] = None
) -> Tuple[Dict[str, Any], Union[List[Suggestion], None], List[VectorSchema], bool, Dict[str, Any]]:
    if caches is None:
        caches = {
            "metadata_properties": {},
            "questions": {},
            "vector_settings": {},
        }

    params = record_update.dict(exclude_unset=True)
    needs_search_engine_update = False
    suggestions = None
    vectors = []

    if "metadata_" in params:
        metadata = params["metadata_"]
        needs_search_engine_update = True
        if metadata is not None:
            caches["metadata_properties"] = await _validate_record_metadata(
                db, dataset, metadata, caches["metadata_properties"]
            )

    if record_update.suggestions is not None:
        params.pop("suggestions")
        questions_ids = [suggestion.question_id for suggestion in record_update.suggestions]
        if len(questions_ids) != len(set(questions_ids)):
            raise ValueError("found duplicate suggestions question IDs")
        suggestions, caches["questions"] = await _build_record_suggestions(db, record_update, caches["questions"])

    if record_update.vectors is not None:
        params.pop("vectors")
        vectors, caches["vector_settings"] = await _build_record_vectors(
            db,
            dataset,
            record_update,
            build_vector_func=lambda value, vector_settings_id: VectorSchema(
                value=value, record_id=record_update.id, vector_settings_id=vector_settings_id
            ),
            cache=caches["vector_settings"],
        )
        needs_search_engine_update = True

    return params, suggestions, vectors, needs_search_engine_update, caches


async def _preload_records_relationships_before_index(db: "AsyncSession", records: List[Record]) -> None:
    for record in records:
        await _preload_record_relationships_before_index(db, record)


async def _preload_record_relationships_before_index(db: "AsyncSession", record: Record) -> None:
    await db.execute(
        select(Record)
        .filter_by(id=record.id)
        .options(
            selectinload(Record.responses).selectinload(Response.user),
            selectinload(Record.suggestions).selectinload(Suggestion.question),
            selectinload(Record.vectors),
        )
    )


async def preload_records_relationships_before_validate(db: "AsyncSession", records: List[Record]) -> None:
    await db.execute(
        select(Record)
        .filter(Record.id.in_([record.id for record in records]))
        .options(
            selectinload(Record.dataset).selectinload(Dataset.questions),
        )
    )


async def update_records(
    db: "AsyncSession", search_engine: "SearchEngine", dataset: Dataset, records_update: "RecordsUpdate"
) -> None:
    records_ids = [record_update.id for record_update in records_update.items]

    if len(records_ids) != len(set(records_ids)):
        raise ValueError("Found duplicate records IDs")

    existing_records_ids = await _exists_records_with_ids(db, dataset_id=dataset.id, records_ids=records_ids)
    non_existing_records_ids = set(records_ids) - set(existing_records_ids)

    if len(non_existing_records_ids) > 0:
        sorted_non_existing_records_ids = sorted(non_existing_records_ids, key=lambda x: records_ids.index(x))
        records_str = ", ".join([str(record_id) for record_id in sorted_non_existing_records_ids])
        raise ValueError(f"Found records that do not exist: {records_str}")

    # Lists to store the records that will be updated in the database or in the search engine
    records_update_objects: List[Dict[str, Any]] = []
    records_search_engine_update: List[UUID] = []
    records_delete_suggestions: List[UUID] = []

    # Cache dictionaries to avoid querying the database multiple times
    caches = {
        "metadata_properties": {},
        "questions": {},
        "vector_settings": {},
    }

    suggestions = []
    upsert_vectors = []
    for record_i, record_update in enumerate(records_update.items):
        try:
            params, record_suggestions, record_vectors, needs_search_engine_update, caches = await _update_record(
                db, dataset, record_update, caches
            )

            if record_suggestions is not None:
                suggestions.extend(record_suggestions)
                records_delete_suggestions.append(record_update.id)

            upsert_vectors.extend(record_vectors)

            if needs_search_engine_update:
                records_search_engine_update.append(record_update.id)

            # Only update the record if there are params to update
            if len(params) > 1:
                records_update_objects.append(params)
        except ValueError as e:
            raise ValueError(f"Record at position {record_i} is not valid because {e}") from e

    async with db.begin_nested():
        if records_delete_suggestions:
            params = [Suggestion.record_id.in_(records_delete_suggestions)]
            await Suggestion.delete_many(db, params=params, autocommit=False)

        if suggestions:
            db.add_all(suggestions)

        if upsert_vectors:
            await Vector.upsert_many(
                db,
                objects=upsert_vectors,
                constraints=[Vector.record_id, Vector.vector_settings_id],
                autocommit=False,
            )

        if records_update_objects:
            await Record.update_many(db, records_update_objects, autocommit=False)

        if records_search_engine_update:
            records = await get_records_by_ids(
                db,
                dataset_id=dataset.id,
                records_ids=records_search_engine_update,
                include=RecordIncludeParam(keys=[RecordInclude.vectors], vectors=None),
            )
            await dataset.awaitable_attrs.vectors_settings
            await _preload_records_relationships_before_index(db, records)
            await search_engine.index_records(dataset, records)

    await db.commit()


async def delete_records(
    db: "AsyncSession", search_engine: "SearchEngine", dataset: Dataset, records_ids: List[UUID]
) -> None:
    async with db.begin_nested():
        params = [Record.id.in_(records_ids), Record.dataset_id == dataset.id]
        records = await Record.delete_many(db=db, params=params, autocommit=False)
        await search_engine.delete_records(dataset=dataset, records=records)

    await db.commit()


async def update_record(
    db: "AsyncSession", search_engine: "SearchEngine", record: Record, record_update: "RecordUpdate"
) -> Record:
    params, suggestions, vectors, needs_search_engine_update, _ = await _update_record(
        db, record.dataset, RecordUpdateWithId(id=record.id, **record_update.dict(by_alias=True, exclude_unset=True))
    )

    # Remove existing suggestions
    if suggestions is not None:
        record.suggestions = []
        params["suggestions"] = suggestions

    async with db.begin_nested():
        record = await record.update(db, **params, replace_dict=True, autocommit=False)

        if vectors:
            await Vector.upsert_many(
                db, objects=vectors, constraints=[Vector.record_id, Vector.vector_settings_id], autocommit=False
            )
            await db.refresh(record, attribute_names=["vectors"])

        if needs_search_engine_update:
            await record.dataset.awaitable_attrs.vectors_settings
            await _preload_record_relationships_before_index(db, record)
            await search_engine.index_records(record.dataset, [record])

    await db.commit()
    return record


async def delete_record(db: "AsyncSession", search_engine: "SearchEngine", record: Record) -> Record:
    async with db.begin_nested():
        record = await record.delete(db=db, autocommit=False)
        await search_engine.delete_records(dataset=record.dataset, records=[record])

    await db.commit()

    return record


async def get_response_by_id(db: "AsyncSession", response_id: UUID) -> Union[Response, None]:
    result = await db.execute(
        select(Response)
        .filter_by(id=response_id)
        .options(selectinload(Response.record).selectinload(Record.dataset).selectinload(Dataset.questions))
    )
    return result.scalar_one_or_none()


async def get_response_by_record_id_and_user_id(
    db: "AsyncSession", record_id: UUID, user_id: UUID
) -> Union[Response, None]:
    result = await db.execute(select(Response).filter_by(record_id=record_id, user_id=user_id))
    return result.scalar_one_or_none()


async def count_responses_by_dataset_id_and_user_id(
    db: "AsyncSession", dataset_id: UUID, user_id: UUID, response_status: Optional[ResponseStatus] = None
) -> int:
    expressions = [Response.user_id == user_id]
    if response_status:
        expressions.append(Response.status == response_status)

    result = await db.execute(
        select(func.count(Response.id))
        .join(Record, and_(Record.id == Response.record_id, Record.dataset_id == dataset_id))
        .filter(*expressions)
    )
    return result.scalar()


async def create_response(
    db: "AsyncSession", search_engine: SearchEngine, record: Record, user: User, response_create: ResponseCreate
) -> Response:
    _validate_response_values(record.dataset, values=response_create.values, status=response_create.status)

    async with db.begin_nested():
        response = await Response.create(
            db,
            values=jsonable_encoder(response_create.values),
            status=response_create.status,
            record_id=record.id,
            user_id=user.id,
            autocommit=False,
        )
        await db.flush([response])
        await _touch_dataset_last_activity_at(db, record.dataset)
        await search_engine.update_record_response(response)

    await db.commit()

    return response


async def update_response(
    db: "AsyncSession", search_engine: SearchEngine, response: Response, response_update: ResponseUpdate
):
    _validate_response_values(response.record.dataset, values=response_update.values, status=response_update.status)

    async with db.begin_nested():
        response = await response.update(
            db,
            values=jsonable_encoder(response_update.values),
            status=response_update.status,
            replace_dict=True,
            autocommit=False,
        )

        await _load_users_from_responses(response)
        await _touch_dataset_last_activity_at(db, response.record.dataset)
        await search_engine.update_record_response(response)

    await db.commit()

    return response


async def upsert_response(
    db: "AsyncSession", search_engine: SearchEngine, record: Record, user: User, response_upsert: ResponseUpsert
) -> Response:
    _validate_response_values(record.dataset, values=response_upsert.values, status=response_upsert.status)

    schema = {
        "values": jsonable_encoder(response_upsert.values),
        "status": response_upsert.status,
        "record_id": response_upsert.record_id,
        "user_id": user.id,
    }

    async with db.begin_nested():
        response = await Response.upsert(
            db,
            schema=schema,
            constraints=[Response.record_id, Response.user_id],
            autocommit=False,
        )

        await _load_users_from_responses(response)
        await _touch_dataset_last_activity_at(db, response.record.dataset)
        await search_engine.update_record_response(response)

    await db.commit()

    return response


async def delete_response(db: "AsyncSession", search_engine: SearchEngine, response: Response) -> Response:
    async with db.begin_nested():
        response = await response.delete(db, autocommit=False)
        await _load_users_from_responses(response)
        await _touch_dataset_last_activity_at(db, response.record.dataset)
        await search_engine.delete_record_response(response)

    await db.commit()

    return response


def _validate_response_values(
    dataset: Dataset,
    values: Union[Dict[str, ResponseValueCreate], Dict[str, ResponseValueUpdate], None],
    status: ResponseStatus,
):
    if not values:
        if status not in [ResponseStatus.discarded, ResponseStatus.draft]:
            raise ValueError("missing response values")
        return

    values_copy = copy.copy(values or {})
    for question in dataset.questions:
        if (
            question.required
            and status == ResponseStatus.submitted
            and not (question.name in values and values_copy.get(question.name))
        ):
            raise ValueError(f"missing question with name={question.name}")

        question_response = values_copy.pop(question.name, None)
        if question_response:
            question.parsed_settings.check_response(question_response, status)

    if values_copy:
        raise ValueError(f"found responses for non configured questions: {list(values_copy.keys())!r}")


def _validate_record_fields(dataset: Dataset, fields: Dict[str, Any]):
    fields_copy = copy.copy(fields or {})
    for field in dataset.fields:
        if field.required and not (field.name in fields_copy and fields_copy.get(field.name) is not None):
            raise ValueError(f"missing required value for field: {field.name!r}")

        value = fields_copy.pop(field.name, None)
        if value and not isinstance(value, str):
            raise ValueError(
                f"wrong value found for field {field.name!r}. Expected {str.__name__!r}, found {type(value).__name__!r}"
            )

    if fields_copy:
        raise ValueError(f"found fields values for non configured fields: {list(fields_copy.keys())!r}")


async def get_suggestion_by_record_id_and_question_id(
    db: "AsyncSession", record_id: UUID, question_id: UUID
) -> Union[Suggestion, None]:
    result = await db.execute(select(Suggestion).filter_by(record_id=record_id, question_id=question_id))
    return result.scalar_one_or_none()


async def _preload_suggestion_relationships_before_index(db: "AsyncSession", suggestion: Suggestion) -> None:
    await db.execute(
        select(Suggestion)
        .filter_by(id=suggestion.id)
        .options(
            selectinload(Suggestion.record).selectinload(Record.dataset),
            selectinload(Suggestion.question),
        )
    )


async def upsert_suggestion(
    db: "AsyncSession",
    search_engine: SearchEngine,
    record: Record,
    question: Question,
    suggestion_create: "SuggestionCreate",
) -> Suggestion:
    question.parsed_settings.check_response(suggestion_create)

    async with db.begin_nested():
        suggestion = await Suggestion.upsert(
            db,
            schema=SuggestionCreateWithRecordId(record_id=record.id, **suggestion_create.dict()),
            constraints=[Suggestion.record_id, Suggestion.question_id],
            autocommit=False,
        )
        await _preload_suggestion_relationships_before_index(db, suggestion)
        await search_engine.update_record_suggestion(suggestion)

    await db.commit()

    return suggestion


async def delete_suggestions(
    db: "AsyncSession", search_engine: SearchEngine, record: Record, suggestions_ids: List[UUID]
) -> None:
    params = [Suggestion.id.in_(suggestions_ids), Suggestion.record_id == record.id]
    suggestions = await list_suggestions_by_id_and_record_id(db, suggestions_ids, record.id)

    async with db.begin_nested():
        await Suggestion.delete_many(db=db, params=params, autocommit=False)
        for suggestion in suggestions:
            await search_engine.delete_record_suggestion(suggestion)

    await db.commit()


async def get_suggestion_by_id(db: "AsyncSession", suggestion_id: "UUID") -> Union[Suggestion, None]:
    result = await db.execute(
        select(Suggestion)
        .filter_by(id=suggestion_id)
        .options(
            selectinload(Suggestion.record).selectinload(Record.dataset),
            selectinload(Suggestion.question),
        )
    )

    return result.scalar_one_or_none()


async def list_suggestions_by_id_and_record_id(
    db: "AsyncSession", suggestion_ids: List[UUID], record_id: UUID
) -> Sequence[Suggestion]:
    result = await db.execute(
        select(Suggestion)
        .filter(Suggestion.record_id == record_id, Suggestion.id.in_(suggestion_ids))
        .options(
            selectinload(Suggestion.record).selectinload(Record.dataset),
            selectinload(Suggestion.question),
        )
    )

    return result.scalars().all()


async def delete_suggestion(db: "AsyncSession", search_engine: SearchEngine, suggestion: Suggestion) -> Suggestion:
    async with db.begin_nested():
        suggestion = await suggestion.delete(db, autocommit=False)
        await search_engine.delete_record_suggestion(suggestion)

    await db.commit()

    return suggestion


async def get_metadata_property_by_id(db: "AsyncSession", metadata_property_id: UUID) -> Optional[MetadataProperty]:
    result = await db.execute(
        select(MetadataProperty).filter_by(id=metadata_property_id).options(selectinload(MetadataProperty.dataset))
    )
    return result.scalar_one_or_none()



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Optional

from starlette.authentication import BaseUser

from argilla_server.security.authentication.claims import Claims


class UserInfo(BaseUser, dict):
    """User info from a provider."""

    @property
    def is_authenticated(self) -> bool:
        return True

    def use_claims(self, claims: Optional[Claims]) -> "UserInfo":
        claims = claims or {}

        for attr, item in claims.items():
            self[attr] = self.__getprop__(item)

        return self

    def __getprop__(self, item, default="") -> Any:
        if callable(item):
            return item(self)
        return self.get(item, default)

    __getattr__ = __getprop__



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from argilla_server.settings import settings

DATASETS_INDEX_NAME = settings.dataset_index_name


def datasets_index_mappings():
    """The index mapping definition for the main datasets index"""
    return {
        "properties": {
            "id": {"type": "keyword"},
            "tags": {
                "type": "nested",
                "properties": {
                    "key": {"type": "keyword"},
                    "value": {"type": "text"},
                },
            },
            "metadata": {
                "type": "nested",
                "properties": {
                    "key": {"type": "keyword"},
                    "value": {"type": "text"},
                },
            },
        }
    }



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from typing import Any, Dict, Optional, TypeVar, Union

from argilla_server.commons.models import TaskType
from argilla_server.constants import ES_INDEX_REGEX_PATTERN
from argilla_server.pydantic_v1 import BaseModel, Field, root_validator


class BaseDatasetDB(BaseModel):
    name: str = Field(regex=ES_INDEX_REGEX_PATTERN)
    task: TaskType
    owner: Optional[str] = Field(description="Deprecated. Use `workspace` instead. Will be removed in v1.5.0")
    workspace: Optional[str]

    tags: Dict[str, str] = Field(default_factory=dict)
    metadata: Dict[str, Any] = Field(default_factory=dict)
    created_at: datetime = None
    created_by: str = Field(
        None,
        description="The argilla user that created the dataset",
    )
    last_updated: datetime = None

    class Config:
        validate_assignment = True

    @root_validator(pre=True)
    def set_defaults(cls, values):
        workspace = values.get("workspace") or values.get("owner")

        cls._check_workspace(workspace)

        values["workspace"] = workspace
        values["owner"] = workspace

        return values

    @classmethod
    def _check_workspace(cls, workspace: str):
        if not workspace:
            raise ValueError("Missing workspace")

    @classmethod
    def build_dataset_id(cls, name: str, workspace: str) -> str:
        """Build a dataset id for a given name and workspace"""
        cls._check_workspace(workspace)
        return f"{workspace}.{name}"

    @property
    def id(self) -> str:
        """The dataset id. Compounded by workspace and name"""
        return self.build_dataset_id(self.name, self.workspace)

    def dict(self, *args, **kwargs) -> Dict[str, Any]:
        """
        Extends base component dict extending object properties
        and user defined extended fields
        """
        return {
            **super().dict(*args, **kwargs),
            "id": self.id,
        }


class EmbeddingsConfig(BaseModel):
    dim: int = Field(
        description="The number of dimensions for the named vectors",
    )


class BaseDatasetSettingsDB(BaseModel):
    vectors: Optional[Dict[str, Union[int, EmbeddingsConfig]]] = Field(
        default=None,
        description="The vectors configuration",
    )


DatasetDB = TypeVar("DatasetDB", bound=BaseDatasetDB)
DatasetSettingsDB = TypeVar("DatasetSettingsDB", bound=BaseDatasetSettingsDB)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from typing import Annotated, Any, Dict, Generic, List, Literal, Optional, TypeVar, Union
from uuid import UUID

from typing_extensions import Annotated

from argilla_server.enums import MetadataPropertyType
from argilla_server.pydantic_v1 import BaseModel, Field, constr, root_validator, validator
from argilla_server.pydantic_v1.generics import GenericModel
from argilla_server.schemas.base import UpdateSchema

FLOAT_METADATA_METRICS_PRECISION = 5

METADATA_PROPERTY_CREATE_NAME_REGEX = r"^(?=.*[a-z0-9])[a-z0-9_-]+$"
METADATA_PROPERTY_CREATE_NAME_MIN_LENGTH = 1
METADATA_PROPERTY_CREATE_NAME_MAX_LENGTH = 200

METADATA_PROPERTY_CREATE_TITLE_MIN_LENGTH = 1
METADATA_PROPERTY_CREATE_TITLE_MAX_LENGTH = 500

TERMS_METADATA_PROPERTY_VALUES_MIN_ITEMS = 1
TERMS_METADATA_PROPERTY_VALUES_MAX_ITEMS = 250

try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated


class TermsMetadataMetrics(BaseModel):
    class TermCount(BaseModel):
        term: str
        count: int

    type: Literal[MetadataPropertyType.terms] = Field(MetadataPropertyType.terms, const=True)
    total: int
    values: List[TermCount] = Field(default_factory=list)


NT = TypeVar("NT", int, float)


class NumericMetadataMetrics(GenericModel, Generic[NT]):
    min: Optional[NT]
    max: Optional[NT]


class IntegerMetadataMetrics(NumericMetadataMetrics[int]):
    type: Literal[MetadataPropertyType.integer] = Field(MetadataPropertyType.integer, const=True)


class FloatMetadataMetrics(NumericMetadataMetrics[float]):
    type: Literal[MetadataPropertyType.float] = Field(MetadataPropertyType.float, const=True)

    @validator("min", "max")
    def round_result(cls, v: float):
        if v is not None:
            return round(v, FLOAT_METADATA_METRICS_PRECISION)
        return v


MetadataMetrics = Annotated[
    Union[TermsMetadataMetrics, IntegerMetadataMetrics, FloatMetadataMetrics], Field(..., discriminator="type")
]


class TermsMetadataProperty(BaseModel):
    type: Literal[MetadataPropertyType.terms]
    values: Optional[List[str]] = None


class IntegerMetadataProperty(BaseModel):
    type: Literal[MetadataPropertyType.integer]
    min: Optional[int] = None
    max: Optional[int] = None


class FloatMetadataProperty(BaseModel):
    type: Literal[MetadataPropertyType.float]
    min: Optional[float] = None
    max: Optional[float] = None


MetadataPropertySettings = Annotated[
    Union[TermsMetadataProperty, IntegerMetadataProperty, FloatMetadataProperty],
    Field(..., discriminator="type"),
]


MetadataPropertyName = Annotated[
    str,
    Field(
        ...,
        regex=METADATA_PROPERTY_CREATE_NAME_REGEX,
        min_length=METADATA_PROPERTY_CREATE_NAME_MIN_LENGTH,
        max_length=METADATA_PROPERTY_CREATE_NAME_MAX_LENGTH,
    ),
]


MetadataPropertyTitle = Annotated[
    constr(min_length=METADATA_PROPERTY_CREATE_TITLE_MIN_LENGTH, max_length=METADATA_PROPERTY_CREATE_TITLE_MAX_LENGTH),
    Field(..., description="The title of the metadata property"),
]


class NumericMetadataProperty(GenericModel, Generic[NT]):
    min: Optional[NT] = None
    max: Optional[NT] = None

    @root_validator(skip_on_failure=True)
    def check_bounds(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        min = values.get("min")
        max = values.get("max")

        if min is not None and max is not None and min >= max:
            raise ValueError(f"'min' ({min}) must be lower than 'max' ({max})")

        return values


class TermsMetadataPropertyCreate(BaseModel):
    type: Literal[MetadataPropertyType.terms]
    values: Optional[List[str]] = Field(
        None, min_items=TERMS_METADATA_PROPERTY_VALUES_MIN_ITEMS, max_items=TERMS_METADATA_PROPERTY_VALUES_MAX_ITEMS
    )


class IntegerMetadataPropertyCreate(NumericMetadataProperty[int]):
    type: Literal[MetadataPropertyType.integer]


class FloatMetadataPropertyCreate(NumericMetadataProperty[float]):
    type: Literal[MetadataPropertyType.float]


MetadataPropertySettingsCreate = Annotated[
    Union[TermsMetadataPropertyCreate, IntegerMetadataPropertyCreate, FloatMetadataPropertyCreate],
    Field(..., discriminator="type"),
]


class MetadataProperty(BaseModel):
    id: UUID
    name: str
    title: str
    settings: MetadataPropertySettings
    visible_for_annotators: bool
    dataset_id: UUID
    inserted_at: datetime
    updated_at: datetime

    class Config:
        orm_mode = True


class MetadataProperties(BaseModel):
    items: List[MetadataProperty]


class MetadataPropertyCreate(BaseModel):
    name: MetadataPropertyName
    title: MetadataPropertyTitle
    settings: MetadataPropertySettingsCreate
    visible_for_annotators: bool = True


class MetadataPropertyUpdate(UpdateSchema):
    title: Optional[MetadataPropertyTitle]
    visible_for_annotators: Optional[bool]

    __non_nullable_fields__ = {"title", "visible_for_annotators"}



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, ClassVar, Dict, Iterable, List, Optional, Set, Tuple

from argilla_server.pydantic_v1 import BaseModel, Field
from argilla_server.services.metrics import ServiceBaseMetric, ServicePythonMetric
from argilla_server.services.metrics.models import CommonTasksMetrics
from argilla_server.services.search.model import ServiceRecordsQuery
from argilla_server.services.tasks.token_classification.model import (
    EntitySpan,
    ServiceTokenClassificationAnnotation,
    ServiceTokenClassificationRecord,
)
from argilla_server.utils import SpanUtils


class F1Metric(ServicePythonMetric[ServiceTokenClassificationRecord]):
    """The F1 metric based on entity-level.

    We follow the convention of `CoNLL 2003 <https://aclanthology.org/W03-0419/>`_, where:
    `"precision is the percentage of named entities found by the learning system that are correct.
    Recall is the percentage of named entities present in the corpus that are found by the system.
    A named entity is correct only if it is an exact match (...).”`
    """

    def apply(self, records: Iterable[ServiceTokenClassificationRecord]) -> Dict[str, Any]:
        # store entities per label in dicts
        predicted_entities = {}
        annotated_entities = {}

        # extract entities per label to dicts
        for rec in records:
            if rec.prediction:
                self._add_entities_to_dict(rec.prediction.entities, predicted_entities)
            if rec.annotation:
                self._add_entities_to_dict(rec.annotation.entities, annotated_entities)

        # store precision, recall, and f1 per label
        per_label_metrics = {}

        annotated_total, predicted_total, correct_total = 0, 0, 0
        precision_macro, recall_macro = 0, 0
        for label, annotated in annotated_entities.items():
            predicted = predicted_entities.get(label, set())
            correct = len(annotated & predicted)

            # safe divides are used to cover the 0/0 cases
            precision = self._safe_divide(correct, len(predicted))
            recall = self._safe_divide(correct, len(annotated))
            per_label_metrics.update(
                {
                    f"{label}_precision": precision,
                    f"{label}_recall": recall,
                    f"{label}_f1": self._safe_divide(2 * precision * recall, precision + recall),
                }
            )

            annotated_total += len(annotated)
            predicted_total += len(predicted)
            correct_total += correct

            precision_macro += precision / len(annotated_entities)
            recall_macro += recall / len(annotated_entities)

        # store macro and micro averaged precision, recall and f1
        averaged_metrics = {
            "precision_macro": precision_macro,
            "recall_macro": recall_macro,
            "f1_macro": self._safe_divide(2 * precision_macro * recall_macro, precision_macro + recall_macro),
        }

        precision_micro = self._safe_divide(correct_total, predicted_total)
        recall_micro = self._safe_divide(correct_total, annotated_total)
        averaged_metrics.update(
            {
                "precision_micro": precision_micro,
                "recall_micro": recall_micro,
                "f1_micro": self._safe_divide(2 * precision_micro * recall_micro, precision_micro + recall_micro),
            }
        )

        return {**averaged_metrics, **per_label_metrics}

    @staticmethod
    def _add_entities_to_dict(entities: List[EntitySpan], dictionary: Dict[str, Set[Tuple[int, int]]]):
        """Helper function for the apply method."""
        for ent in entities:
            try:
                dictionary[ent.label].add((ent.start, ent.end))
            except KeyError:
                dictionary[ent.label] = {(ent.start, ent.end)}

    @staticmethod
    def _safe_divide(numerator, denominator):
        """Helper function for the apply method."""
        try:
            return numerator / denominator
        except ZeroDivisionError:
            return 0


class DatasetLabels(ServicePythonMetric):
    id: str = Field("dataset_labels", const=True)
    name: str = Field("The dataset entity labels", const=True)
    records_to_fetch: int = 1500

    shuffle_records: bool = Field(default=True)

    def prepare_query(self, query: ServiceRecordsQuery):
        text = query.query_text or ""
        if text:
            text += " AND "
        text += " (_exists_:annotated_by OR _exists_:predicted_by)"

        query.query_text = text
        return query

    def apply(
        self,
        records: Iterable[ServiceTokenClassificationRecord],
    ) -> Dict[str, Any]:
        ds_labels = set()

        for _ in range(0, self.records_to_fetch):  # Only a few of records will be parsed
            record: ServiceTokenClassificationRecord = next(records, None)
            if record is None:
                break

            if record.annotation:
                ds_labels.update([entity.label for entity in record.annotation.entities])
            if record.prediction:
                ds_labels.update([entity.label for entity in record.prediction.entities])
        return {"labels": ds_labels or []}


class MentionMetrics(BaseModel):
    """Mention metrics model"""

    value: str
    label: str
    score: float = Field(ge=0.0)
    capitalness: Optional[str] = Field(None)


class TokenMetrics(BaseModel):
    """
    Token metrics stored in elasticsearch for token classification

    Attributes
        idx: The token index in sentence
        value: The token textual value
        char_start: The token character start position in sentence
        char_end: The token character end position in sentence
        score: Token score info
        tag: Token tag info. Deprecated: Use metrics.predicted.tags or metrics.annotated.tags instead
        custom: extra token level info
    """

    value: str
    capitalness: Optional[str] = None


class TokenClassificationMetrics(CommonTasksMetrics[ServiceTokenClassificationRecord]):
    """Configured metrics for token classification"""

    @staticmethod
    def density(value: int, sentence_length: int) -> float:
        """Compute the string density over a sentence"""
        return value / sentence_length

    @staticmethod
    def capitalness(value: str) -> Optional[str]:
        """Compute capitalness for a string value"""
        value = value.strip()
        if not value:
            return None
        if value.isupper():
            return "UPPER"
        if value.islower():
            return "LOWER"
        if value[0].isupper():
            return "FIRST"
        if any([c.isupper() for c in value[1:]]):
            return "MIDDLE"
        return None

    @staticmethod
    def mentions_metrics(record: ServiceTokenClassificationRecord, mentions: List[Tuple[str, EntitySpan]]):
        def mention_tokens_length(entity: EntitySpan) -> int:
            """Compute mention tokens length"""
            return len(
                set(
                    [
                        token_idx
                        for i in range(entity.start, entity.end)
                        for token_idx in [record.span_utils.char_to_token_idx.get(i)]
                        if token_idx is not None
                    ]
                )
            )

        return [
            MentionMetrics(
                value=mention,
                label=entity.label,
                score=entity.score,
                capitalness=TokenClassificationMetrics.capitalness(mention),
            )
            for mention, entity in mentions
        ]

    @classmethod
    def build_tokens_metrics(cls, record: ServiceTokenClassificationRecord) -> List[TokenMetrics]:
        return [
            TokenMetrics(value=token_value, capitalness=cls.capitalness(token_value))
            for token_idx, token_value in enumerate(record.tokens)
        ]

    @classmethod
    def record_metrics(cls, record: ServiceTokenClassificationRecord) -> Dict[str, Any]:
        """Compute metrics at record level"""
        base_metrics = super(TokenClassificationMetrics, cls).record_metrics(record)

        tokens_metrics = cls.build_tokens_metrics(record)
        return {
            **base_metrics,
            "tokens": tokens_metrics,
            "predicted": {"mentions": cls.mentions_metrics(record, record.predicted_mentions())},
            "annotated": {"mentions": cls.mentions_metrics(record, record.annotated_mentions())},
        }

    @staticmethod
    def _compute_iob_tags(
        span_utils: SpanUtils,
        annotation: Optional[ServiceTokenClassificationAnnotation],
    ) -> Optional[List[str]]:
        """Helper method to compute IOB tags from entity spans

        Args:
            span_utils: Helper class to perform the computation.
            annotation: Contains the spans from which to compute the IOB tags.

        Returns:
            The IOB tags or None if ``annotation`` is None.
        """
        if annotation is None:
            return None

        spans = [(ent.label, ent.start, ent.end) for ent in annotation.entities]
        return span_utils.to_tags(spans)

    metrics: ClassVar[List[ServiceBaseMetric]] = (
        CommonTasksMetrics.metrics
        + [
            DatasetLabels(),
            F1Metric(
                id="F1",
                name="F1 ServiceBaseMetric based on entity-level",
                description="F1 metrics based on entity-level (averaged and per label), "
                "where only exact matches count (CoNNL 2003).",
            ),
        ]
        + [
            ServiceBaseMetric(
                id="predicted_as",
                name="Predicted labels distribution",
            ),
            ServiceBaseMetric(
                id="annotated_as",
                name="Annotated labels distribution",
            ),
            ServiceBaseMetric(
                id="token_frequency",
                name="Tokens frequency distribution",
            ),
            ServiceBaseMetric(
                id="token_capitalness",
                name="Token capitalness distribution",
                description="Computes capitalization information of tokens",
            ),
            ServiceBaseMetric(
                id="predicted_entity_labels",
                name="Predicted entity labels",
                description="Predicted entity labels distribution",
            ),
            ServiceBaseMetric(
                id="predicted_entity_capitalness",
                name="Mention entity capitalness for predictions",
                description="Computes capitalization information of predicted entity mentions",
            ),
            ServiceBaseMetric(
                id="predicted_mentions_distribution",
                name="Predicted mentions distribution by entity",
                description="Computes predicted mentions distribution against its labels",
            ),
            ServiceBaseMetric(
                id="predicted_top_k_mentions_consistency",
                name="Entity label consistency for predictions",
                description="Computes entity label variability for top-k predicted entity mentions",
            ),
            ServiceBaseMetric(
                id="annotated_entity_labels",
                name="Annotated entity labels",
                description="Annotated Entity labels distribution",
            ),
            ServiceBaseMetric(
                id="annotated_entity_capitalness",
                name="Mention entity capitalness for annotations",
                description="Compute capitalization information of annotated entity mentions",
            ),
            ServiceBaseMetric(
                id="annotated_mentions_distribution",
                name="Annotated mentions distribution by entity",
                description="Computes annotated mentions distribution against its labels",
            ),
            ServiceBaseMetric(
                id="annotated_top_k_mentions_consistency",
                name="Entity label consistency for annotations",
                description="Computes entity label variability for top-k annotated entity mentions",
            ),
            # TODO: Remove in v1.10.0
            ServiceBaseMetric(
                id="tokens_length",
                name="Tokens length",
                description="Computes the text length distribution measured in number of tokens",
            ),
            # TODO: Remove in v1.10.0
            ServiceBaseMetric(
                id="token_length",
                name="Token length distribution",
                description="Computes token length distribution in number of characters",
            ),
            # TODO: Remove in v1.10.0
            ServiceBaseMetric(
                id="predicted_entity_density",
                name="Mention entity density for predictions",
                description="Computes the ratio between the number of all entity tokens and tokens in the text",
            ),
            # TODO: Remove in v1.10.0
            ServiceBaseMetric(
                id="predicted_mention_token_length",
                name="Predicted mention tokens length",
                description="Computes the length of the predicted entity mention measured in number of tokens",
            ),
            # TODO: Remove in v1.10.0
            ServiceBaseMetric(
                id="predicted_mention_char_length",
                name="Predicted mention characters length",
                description="Computes the length of the predicted entity mention measured in number of tokens",
            ),
            # TODO: Remove in v1.10.0
            ServiceBaseMetric(
                id="predicted_tag_consistency",
                name="Token tag consistency for predictions",
                description="Computes token tag variability for top-k predicted tags",
            ),
            # TODO: Remove in v1.10.0
            ServiceBaseMetric(
                id="annotated_entity_density",
                name="Mention entity density for annotations",
                description="Computes the ratio between the number of all entity tokens and tokens in the text",
            ),
            # TODO: Remove in v1.10.0
            ServiceBaseMetric(
                id="annotated_mention_token_length",
                name="Annotated mention tokens length",
                description="Computes the length of the entity mention measured in number of tokens",
            ),
            # TODO: Remove in v1.10.0
            ServiceBaseMetric(
                id="annotated_mention_char_length",
                name="Annotated mention characters length",
                description="Computes the length of the entity mention measured in number of tokens",
            ),
            # TODO: Remove in v1.10.0
            ServiceBaseMetric(
                id="annotated_tag_consistency",
                name="Token tag consistency for annotations",
                description="Computes token tag variability for top-k annotated tags",
            ),
        ]
    )



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Any, Dict, List

import pytest
from argilla_server.apis.v0.models.text_classification import TextClassificationBulkRequest
from argilla_server.commons.models import TaskType
from argilla_server.constants import API_KEY_HEADER_NAME
from argilla_server.models import UserRole
from argilla_server.schemas.v0.datasets import Dataset

from tests.factories import (
    AnnotatorFactory,
    DatasetFactory,
    UserFactory,
    WorkspaceFactory,
)


@pytest.mark.asyncio
class TestSuiteDatasetApi:
    async def delete_dataset(
        self, async_client: "AsyncClient", dataset_name: str, workspace: str, headers: Dict[str, Any]
    ):
        url = f"/api/datasets/{dataset_name}?workspace={workspace}"

        response = await async_client.delete(url, headers=headers)
        assert response.status_code == 200

    async def create_mock_dataset(
        self,
        async_client: "AsyncClient",
        dataset_name: str,
        headers: Dict[str, Any],
        workspace: str,
        records: List = None,
    ):
        response = await async_client.post(
            "/api/datasets",
            json={"name": dataset_name, "workspace": workspace, "task": TaskType.text_classification.value},
            headers=headers,
        )
        assert response.status_code == 200

        url = f"/api/datasets/{dataset_name}/TextClassification:bulk?workspace={workspace}"

        bulk_data = TextClassificationBulkRequest(
            tags={"env": "test", "class": "text classification"},
            metadata={"config": {"the": "config"}},
            records=records or [],
        ).dict(by_alias=True)

        response = await async_client.post(url, json=bulk_data, headers=headers)
        assert response.status_code == 200, response.json()

    @pytest.mark.parametrize("role", [UserRole.owner, UserRole.admin])
    async def test_delete_dataset(self, async_client: "AsyncClient", owner_auth_header: dict, role: UserRole):
        dataset = await DatasetFactory.create()
        user = await UserFactory(role=role, workspaces=[dataset.workspace])
        api_auth_headers = {API_KEY_HEADER_NAME: user.api_key}

        await self.create_mock_dataset(
            async_client, dataset_name=dataset.name, workspace=dataset.workspace.name, headers=owner_auth_header
        )

        await self.delete_dataset(
            async_client, dataset_name=dataset.name, workspace=dataset.workspace.name, headers=api_auth_headers
        )
        response = await async_client.get(
            f"/api/datasets/{dataset.name}?workspace={dataset.workspace.name}", headers=api_auth_headers
        )

        assert response.status_code == 404, response.json()

        assert response.json() == {
            "detail": {
                "code": "argilla.api.errors::EntityNotFoundError",
                "params": {"name": dataset.name, "type": "ServiceDataset"},
            }
        }

    async def test_delete_dataset_with_missing_workspace(self, async_client: "AsyncClient", owner_auth_header: dict):
        response = await async_client.delete("/api/datasets/dataset", headers=owner_auth_header)

        assert response.status_code == 400
        assert response.json() == {
            "detail": {
                "code": "argilla.api.errors::MissingInputParamError",
                "params": {"message": "A workspace must be provided"},
            }
        }

    @pytest.mark.parametrize("role", [UserRole.owner, UserRole.admin])
    async def test_create_dataset(self, async_client: "AsyncClient", owner_auth_header: dict, role: UserRole):
        dataset = await DatasetFactory.create()
        workspace_name = dataset.workspace.name
        await self.delete_dataset(
            async_client, dataset_name=dataset.name, workspace=workspace_name, headers=owner_auth_header
        )

        user = await UserFactory.create(role=role, workspaces=[dataset.workspace])

        request = dict(
            name=dataset.name,
            task=TaskType.text_classification,
            tags={"env": "test", "class": "text classification"},
            metadata={"config": {"the": "config"}},
        )

        response = await async_client.post(
            f"/api/datasets?workspace={workspace_name}", json=request, headers={API_KEY_HEADER_NAME: user.api_key}
        )
        assert response.status_code == 200

        created_dataset = Dataset.parse_obj(response.json())

        assert created_dataset.id
        assert created_dataset.created_by == user.username
        assert created_dataset.metadata == request["metadata"]
        assert created_dataset.tags == request["tags"]
        assert created_dataset.name == dataset.name
        assert created_dataset.workspace == workspace_name
        assert created_dataset.task == TaskType.text_classification

    async def test_create_dataset_with_already_created_error(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ):
        workspace = await WorkspaceFactory.create()
        dataset = await DatasetFactory.create()

        await self.create_mock_dataset(
            async_client, dataset_name=dataset.name, headers=owner_auth_header, workspace=workspace.name
        )

        response = await async_client.post(
            f"/api/datasets?workspace={workspace.name}",
            json={"name": dataset.name, "task": TaskType.text_classification},
            headers=owner_auth_header,
        )
        assert response.status_code == 409

    async def test_create_workspace_with_missing_workspace(self, async_client: "AsyncClient", owner_auth_header: dict):
        dataset_name = "missing-workspace-dataset"

        response = await async_client.post(
            "/api/datasets",
            json={"name": dataset_name, "task": TaskType.text_classification},
            headers=owner_auth_header,
        )

        assert response.status_code == 400

    @pytest.mark.parametrize("role", [UserRole.owner, UserRole.admin])
    async def test_create_dataset_using_several_workspaces(self, async_client: "AsyncClient", role: UserRole):
        workspaces = await WorkspaceFactory.create_batch(3)
        user = await UserFactory.create(workspaces=workspaces, role=role)
        user_auth_headers = {API_KEY_HEADER_NAME: user.api_key}

        dataset_name = "test_dataset"

        other_workspace = user.workspaces[0]
        for workspace in user.workspaces[1:]:
            await self.delete_dataset(
                async_client, dataset_name=dataset_name, workspace=workspace.name, headers=user_auth_headers
            )
            await self.delete_dataset(
                async_client, dataset_name=dataset_name, workspace=other_workspace.name, headers=user_auth_headers
            )

            request = dict(name=dataset_name, task=TaskType.text_classification)
            response = await async_client.post(
                f"/api/datasets?workspace={workspace.name}", json=request, headers=user_auth_headers
            )

            assert response.status_code == 200, response.json()

            dataset = Dataset.parse_obj(response.json())

            assert dataset.created_by == user.username
            assert dataset.name == dataset_name
            assert dataset.owner == workspace.name
            assert dataset.task == TaskType.text_classification

            response = await async_client.post(
                f"/api/datasets?workspace={workspace.name}", json=request, headers=user_auth_headers
            )

            assert response.status_code == 409, response.json()

            response = await async_client.post(
                f"/api/datasets?workspace={other_workspace.name}", json=request, headers=user_auth_headers
            )

            assert response.status_code == 200, response.json()

            dataset = Dataset.parse_obj(response.json())

            assert dataset.created_by == user.username
            assert dataset.name == dataset_name
            assert dataset.workspace == other_workspace.name
            assert dataset.task == TaskType.text_classification

    @pytest.mark.parametrize("task", [TaskType.text_classification, TaskType.token_classification, TaskType.text2text])
    async def test_dataset_naming_validation(self, async_client: "AsyncClient", owner_auth_header: dict, task):
        dataset_name = "Wrong dataset name"
        workspace = await WorkspaceFactory.create()

        request = {"name": dataset_name, "task": task.value, "workspace": workspace.name}
        response = await async_client.post("/api/datasets", json=request, headers=owner_auth_header)

        assert response.status_code == 422
        assert response.json() == {
            "detail": {
                "code": "argilla.api.errors::ValidationError",
                "params": {
                    "errors": [
                        {
                            "ctx": {"pattern": "^(?!-|_)[a-z0-9-_]+$"},
                            "loc": ["body", "name"],
                            "msg": "string does not match regex " '"^(?!-|_)[a-z0-9-_]+$"',
                            "type": "value_error.str.regex",
                        }
                    ],
                },
            }
        }

    async def test_list_datasets(self, async_client: "AsyncClient", owner_auth_header: dict):
        workspace = await WorkspaceFactory.create()
        dataset = await DatasetFactory.create()
        await self.create_mock_dataset(
            async_client, dataset_name=dataset.name, headers=owner_auth_header, workspace=workspace.name
        )

        response = await async_client.get(f"/api/datasets?workspace={workspace.name}", headers=owner_auth_header)
        assert response.status_code == 200

        datasets = [Dataset.parse_obj(item) for item in response.json()]
        assert len(datasets) > 0

        assert dataset.name in [ds.name for ds in datasets]
        for ds in datasets:
            assert ds.id

    @pytest.mark.parametrize("role", [UserRole.owner, UserRole.admin])
    async def test_list_datasets_without_workspace(
        self, async_client: "AsyncClient", owner_auth_header: dict, role: UserRole
    ):
        dataset = await DatasetFactory.create()
        user = await UserFactory.create(role=role, workspaces=[dataset.workspace])

        await self.create_mock_dataset(
            async_client, dataset.name, headers=owner_auth_header, workspace=dataset.workspace.name
        )
        response = await async_client.get("/api/datasets", headers={API_KEY_HEADER_NAME: user.api_key})

        assert response.status_code == 200

    @pytest.mark.parametrize("role", [UserRole.owner, UserRole.admin, UserRole.annotator])
    async def test_update_dataset(self, async_client: "AsyncClient", owner_auth_header: dict, role: UserRole):
        dataset = await DatasetFactory.create()
        workspace_name = dataset.workspace.name
        await self.create_mock_dataset(
            async_client, dataset_name=dataset.name, headers=owner_auth_header, workspace=workspace_name
        )

        user = await UserFactory.create(role=role, workspaces=[dataset.workspace])
        response = await async_client.patch(
            f"/api/datasets/{dataset.name}?workspace={workspace_name}",
            json={"metadata": {"new": "value"}},
            headers={API_KEY_HEADER_NAME: user.api_key},
        )
        assert response.status_code == 200

        response = await async_client.get(
            f"/api/datasets/{dataset.name}?workspace={workspace_name}", headers={API_KEY_HEADER_NAME: user.api_key}
        )
        assert response.status_code == 200, response.json()

        ds = Dataset.parse_obj(response.json())
        assert ds.metadata["new"] == "value"

    @pytest.mark.parametrize("role", [UserRole.admin, UserRole.annotator])
    async def test_update_dataset_without_permissions(
        self, async_client: "AsyncClient", owner_auth_header: dict, role: UserRole
    ):
        dataset = await DatasetFactory.create()
        workspace_name = dataset.workspace.name
        await self.create_mock_dataset(
            async_client, dataset_name=dataset.name, headers=owner_auth_header, workspace=workspace_name
        )

        user = await UserFactory.create(role=role)

        response = await async_client.patch(
            f"/api/datasets/{dataset.name}?workspace={workspace_name}",
            json={"metadata": {"new": "metadata"}},
            headers={API_KEY_HEADER_NAME: user.api_key},
        )

        assert response.status_code == 403

    @pytest.mark.parametrize("role", [UserRole.admin, UserRole.annotator])
    async def test_update_dataset_by_restricted_user_without_changes(
        self, async_client: "AsyncClient", owner_auth_header: dict, role: UserRole
    ):
        dataset = await DatasetFactory.create()
        workspace_name = dataset.workspace.name
        await self.create_mock_dataset(
            async_client, dataset_name=dataset.name, headers=owner_auth_header, workspace=workspace_name
        )

        user = await UserFactory.create(workspaces=[dataset.workspace], role=role)

        response = await async_client.get(
            f"/api/datasets/{dataset.name}?workspace={workspace_name}", headers={API_KEY_HEADER_NAME: user.api_key}
        )
        assert response.status_code == 200

        created_dataset = response.json()
        created_dataset.pop("last_updated")

        response = await async_client.patch(
            f"/api/datasets/{dataset.name}?workspace={workspace_name}",
            json={},
            headers={API_KEY_HEADER_NAME: user.api_key},
        )

        assert response.status_code == 200

        updated_dataset = response.json()
        updated_dataset.pop("last_updated")

        assert updated_dataset == created_dataset

    @pytest.mark.parametrize("role", [UserRole.owner, UserRole.admin])
    async def test_open_and_close_dataset(self, async_client: "AsyncClient", owner_auth_header: dict, role: UserRole):
        dataset = await DatasetFactory.create()
        workspace_name = dataset.workspace.name
        await self.create_mock_dataset(
            async_client, dataset_name=dataset.name, headers=owner_auth_header, workspace=workspace_name
        )

        user = await UserFactory.create(role=role, workspaces=[dataset.workspace])
        endpoint_url = f"/api/datasets/{dataset.name}:{{action}}?workspace={workspace_name}"

        response = await async_client.put(endpoint_url.format(action="close"), headers=owner_auth_header)
        assert response.status_code == 200

        response = await async_client.post(
            f"/api/datasets/{dataset.name}/TextClassification:search?workspace={workspace_name}",
            headers={API_KEY_HEADER_NAME: user.api_key},
        )
        assert response.status_code == 400
        assert response.json() == {
            "detail": {"code": "argilla.api.errors::ClosedDatasetError", "params": {"name": dataset.name}}
        }

        response = await async_client.put(
            endpoint_url.format(action="open"), headers={API_KEY_HEADER_NAME: user.api_key}
        )
        assert response.status_code == 200

        response = await async_client.post(
            f"/api/datasets/{dataset.name}/TextClassification:search?workspace={workspace_name}",
            headers=owner_auth_header,
        )
        assert response.status_code == 200

    @pytest.mark.parametrize("role", [UserRole.owner, UserRole.admin])
    async def test_delete_records(self, async_client: "AsyncClient", owner_auth_header: dict, role: UserRole):
        dataset = await DatasetFactory.create()
        records = [{"id": i, "inputs": {"text": f"This is a text for id {i}"}} for i in range(1, 100)]
        workspace_name = dataset.workspace.name
        await self.create_mock_dataset(
            async_client,
            dataset_name=dataset.name,
            workspace=workspace_name,
            headers=owner_auth_header,
            records=records,
        )

        user = await UserFactory.create(role=role, workspaces=[dataset.workspace])

        response = await async_client.request(
            "DELETE",
            url=f"/api/datasets/{dataset.name}/data?workspace={workspace_name}",
            json={"ids": [1]},
            headers={API_KEY_HEADER_NAME: user.api_key},
        )

        assert response.status_code == 200
        assert response.json() == {"matched": 1, "processed": 1}

        response = await async_client.delete(
            f"/api/datasets/{dataset.name}/data?mark_as_discarded=true&workspace={workspace_name}",
            headers={API_KEY_HEADER_NAME: user.api_key},
        )

        assert response.status_code == 200
        assert response.json() == {"matched": 99, "processed": 98}  # different values are caused by conflicts found

    async def test_delete_records_as_annotator(self, async_client: "AsyncClient", owner_auth_header: dict):
        dataset = await DatasetFactory.create()
        records = [{"id": i, "inputs": {"text": f"This is a text for id {i}"}} for i in range(1, 100)]
        workspace_name = dataset.workspace.name
        await self.create_mock_dataset(
            async_client,
            dataset_name=dataset.name,
            workspace=workspace_name,
            headers=owner_auth_header,
            records=records,
        )

        user = await AnnotatorFactory.create(workspaces=[dataset.workspace])

        response = await async_client.delete(
            f"/api/datasets/{dataset.name}/data?workspace={workspace_name}",
            headers={API_KEY_HEADER_NAME: user.api_key},
        )

        assert response.status_code == 403
        assert response.json() == {
            "detail": {
                "code": "argilla.api.errors::ForbiddenOperationError",
                "params": {
                    "detail": "You don't have the necessary permissions to delete records on this dataset."
                    " Only administrators can delete datasets"
                },
            }
        }

        response = await async_client.delete(
            f"/api/datasets/{dataset.name}/data",
            headers={API_KEY_HEADER_NAME: user.api_key},
            params={
                "mark_as_discarded": True,
                "workspace": workspace_name,
            },
        )

        assert response.status_code == 200
        assert response.json() == {"matched": 99, "processed": 99}  # different values are caused by conflicts found



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

import psutil


def server_deployment_type() -> str:
    """Returns the type of deployment of the server."""
    global _server_deployment_type

    if _server_deployment_type is None:
        _server_deployment_type = "server"

        if is_running_on_docker_container() and _is_quickstart_env():
            _server_deployment_type = "quickstart"

    return _server_deployment_type


def is_running_on_docker_container() -> bool:
    """Returns True if the current process is running inside a Docker container, False otherwise."""
    global _in_docker_container

    if _in_docker_container is None:
        _in_docker_container = _has_docker_env() or _has_docker_cgroup()

    return _in_docker_container


def _has_docker_env() -> bool:
    try:
        return os.path.exists("/.dockerenv")
    except:
        return False


def _has_docker_cgroup() -> bool:
    try:
        cgroup_path = "/proc/self/cgroup"
        return (
            os.path.exists(cgroup_path)
            and os.path.isfile(cgroup_path)
            and any("docker" in line for line in open(cgroup_path))
        )
    except:
        return False


def _is_quickstart_script_running() -> bool:
    # TODO: Any modification in the `quickstart.Dockerfile` file should be reflected here

    try:
        cmdline_key = "cmdline"
        for process in psutil.process_iter([cmdline_key]):
            if "start_quickstart_argilla.sh" in process.info[cmdline_key]:
                return True
    except:
        pass
    return False


def _is_quickstart_env():
    # TODO: Any modification in the `quickstart.Dockerfile` file should be reflected here

    for env_var in [
        "OWNER_USERNAME",
        "OWNER_PASSWORD",
        "OWNER_API_KEY",
        "ADMIN_USERNAME",
        "ADMIN_PASSWORD",
        "ADMIN_API_KEY",
        "ANNOTATOR_USERNAME",
        "ANNOTATOR_PASSWORD",
        "LOAD_DATASETS",
    ]:
        if env_var not in os.environ:
            return False
    return True


# Private global variables section
_in_docker_container = None
_server_deployment_type = None



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import TYPE_CHECKING

import pytest
from argilla_server.apis.v0.models.text_classification import (
    CreateLabelingRule,
    LabelingRule,
    LabelingRuleMetricsSummary,
    TextClassificationBulkRequest,
    TextClassificationRecord,
)
from argilla_server.commons.models import TaskType
from argilla_server.constants import API_KEY_HEADER_NAME
from argilla_server.models import User

if TYPE_CHECKING:
    from httpx import AsyncClient


async def log_some_records(
    client: "AsyncClient",
    dataset: str,
    workspace_name: str,
    annotation: str = None,
    multi_label: bool = False,
    delete: bool = True,
):
    if delete:
        response = await client.delete(f"/api/datasets/{dataset}", params={"workspace": workspace_name})
        assert response.status_code == 200

        response = await client.post(
            "/api/datasets",
            json={"name": dataset, "task": TaskType.text_classification.value, "workspace": workspace_name},
        )
        assert response.status_code == 200, response.json()

    record = {
        "id": 0,
        "inputs": {"text": "Esto es un ejemplo de texto"},
        "metadata": {"field.one": 1, "field.two": 2},
        "multi_label": multi_label,
    }

    if annotation:
        record["annotation"] = {"agent": "test", "labels": [{"class": annotation}]}

    response = await client.post(
        f"/api/datasets/{dataset}/TextClassification:bulk",
        json=TextClassificationBulkRequest(records=[TextClassificationRecord(**record)]).dict(by_alias=True),
        params={"workspace": workspace_name},
    )
    assert response.status_code == 200, response.json()


@pytest.mark.asyncio
async def test_dataset_without_rules(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_dataset_without_rules"
    await log_some_records(async_client, dataset, workspace_name=argilla_user.username)

    response = await async_client.get(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules", params=workspace_query_params
    )
    assert response.status_code == 200
    assert len(response.json()) == 0


@pytest.mark.asyncio
async def test_dataset_update_rule(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_dataset_with_rules"
    query = "a query"
    await log_some_records(async_client, dataset, workspace_name=argilla_user.username)

    response = await async_client.post(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules",
        json=CreateLabelingRule(query=query, label="LALA").dict(),
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = await async_client.patch(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules/{query}",
        json={"label": "NEW Label"},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = await async_client.get(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules", params=workspace_query_params
    )
    assert response.status_code == 200

    rules = list(map(LabelingRule.parse_obj, response.json()))
    assert len(rules) == 1
    assert rules[0].label == "NEW Label"
    assert rules[0].labels == ["NEW Label"]
    assert rules[0].description is None

    response = await async_client.patch(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules/{query}",
        json={"labels": ["A", "B"], "description": "New description"},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = await async_client.get(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules", params=workspace_query_params
    )
    assert response.status_code == 200

    rules = list(map(LabelingRule.parse_obj, response.json()))
    assert len(rules) == 1
    assert rules[0].description == "New description"
    assert rules[0].label is None
    assert rules[0].labels == ["A", "B"]


@pytest.mark.parametrize(
    "rule",
    [
        CreateLabelingRule(query="a query", description="Description", label="LALA"),
        CreateLabelingRule(query="/a qu?ry/", description="Description", label="LALA"),
        CreateLabelingRule(query="another query", description="Description", labels=["A", "B", "C"]),
    ],
)
@pytest.mark.asyncio
async def test_dataset_with_rules(async_client: "AsyncClient", argilla_user: User, rule: CreateLabelingRule):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_dataset_with_rules"
    await log_some_records(async_client, dataset, workspace_name=argilla_user.username)

    response = await async_client.post(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules", json=rule.dict(), params=workspace_query_params
    )
    assert response.status_code == 200

    created_rule = LabelingRule.parse_obj(response.json())
    assert created_rule.query == rule.query
    assert created_rule.label == rule.label
    assert created_rule.labels == rule.labels
    assert created_rule.description == rule.description

    response = await async_client.get(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules", params=workspace_query_params
    )
    assert response.status_code == 200

    rules = list(map(LabelingRule.parse_obj, response.json()))
    assert len(rules) == 1
    assert rules[0] == created_rule


@pytest.mark.parametrize(
    "rule",
    [
        CreateLabelingRule(query="a query", description="Description", label="LALA"),
        CreateLabelingRule(query="/a qu(e|E)ry/", description="Description", label="LALA"),
        CreateLabelingRule(query="another query", description="Description", labels=["A", "B", "C"]),
    ],
)
@pytest.mark.asyncio
async def test_get_dataset_rule(async_client: "AsyncClient", argilla_user: User, rule: CreateLabelingRule):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_get_dataset_rule"
    await log_some_records(async_client, dataset, workspace_name=argilla_user.username)

    response = await async_client.post(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules",
        json=rule.dict(),
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = await async_client.get(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules/{rule.query}", params=workspace_query_params
    )
    assert response.status_code == 200

    found_rule = LabelingRule.parse_obj(response.json())
    assert found_rule.query == rule.query
    assert found_rule.label == rule.label
    assert found_rule.labels == rule.labels
    assert found_rule.description == rule.description


@pytest.mark.asyncio
async def test_delete_dataset_rules(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_delete_dataset_rules"
    await log_some_records(async_client, dataset, workspace_name=argilla_user.username)

    response = await async_client.post(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules",
        json=CreateLabelingRule(query="/a query/", label="TEST", description="Description").dict(),
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = await async_client.delete(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules//a query/", params=workspace_query_params
    )
    assert response.status_code == 200

    response = await async_client.get(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules", params=workspace_query_params
    )
    assert response.status_code == 200

    assert len(response.json()) == 0


@pytest.mark.asyncio
async def test_duplicated_dataset_rules(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_duplicated_dataset_rules"
    await log_some_records(async_client, dataset, workspace_name=argilla_user.username)

    rule = CreateLabelingRule(query="a query", labels=["TEST"])
    response = await async_client.post(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules",
        json=rule.dict(),
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules", json=rule.dict(), params=workspace_query_params
    )
    assert response.status_code == 409
    assert response.json() == {
        "detail": {
            "code": "argilla.api.errors::EntityAlreadyExistsError",
            "params": {"name": "a query", "type": "ServiceLabelingRule"},
        }
    }


@pytest.mark.asyncio
async def test_rules_with_multi_label_dataset(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_rules_with_multi_label_dataset"
    await log_some_records(async_client, dataset, workspace_name=argilla_user.username, multi_label=True)

    rule = CreateLabelingRule(query="a query", description="Description", label="LALA")
    response = await async_client.post(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules", json=rule.dict(), params=workspace_query_params
    )
    assert response.status_code == 200

    found_rule = LabelingRule.parse_obj(response.json())
    assert found_rule.query == rule.query
    assert found_rule.label == rule.label
    assert found_rule.labels == rule.labels
    assert found_rule.description == rule.description


@pytest.mark.asyncio
async def test_rule_metrics_with_missing_label(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_rule_metrics_with_missing_label"
    await log_some_records(async_client, dataset, workspace_name=argilla_user.username, annotation="OK")

    response = await async_client.get(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules/a query/metrics", params=workspace_query_params
    )
    assert response.status_code == 200, response.json()

    assert response.json() == {
        "coverage": 0.0,
        "coverage_annotated": 0.0,
        "correct": 0.0,
        "incorrect": 0.0,
        "total_records": 1,
        "annotated_records": 1,
    }


@pytest.mark.parametrize(
    ("rule", "expected_metrics"),
    [
        (
            CreateLabelingRule(query="ejemplo", label="TEST"),
            {
                "annotated_records": 1,
                "correct": 0.0,
                "coverage": 1.0,
                "coverage_annotated": 1.0,
                "incorrect": 1.0,
                "precision": 0.0,
                "total_records": 1,
            },
        ),
        (
            CreateLabelingRule(query="ejemplo", label="test.bad"),
            {
                "annotated_records": 1,
                "correct": 0.0,
                "coverage": 1.0,
                "coverage_annotated": 1.0,
                "incorrect": 1.0,
                "precision": 0.0,
                "total_records": 1,
            },
        ),
        (
            CreateLabelingRule(query="ejemplo", label="o.k."),
            {
                "annotated_records": 1,
                "correct": 1.0,
                "coverage": 1.0,
                "coverage_annotated": 1.0,
                "incorrect": 0.0,
                "precision": 1.0,
                "total_records": 1,
            },
        ),
        (
            CreateLabelingRule(query="bad query", label="TEST"),
            {
                "annotated_records": 1,
                "correct": 0.0,
                "coverage": 0.0,
                "coverage_annotated": 0.0,
                "incorrect": 0.0,
                "total_records": 1,
            },
        ),
        (
            CreateLabelingRule(query="ejemplo", labels=["TEST", "A", "B"]),
            {
                "annotated_records": 1,
                "correct": 0.0,
                "coverage": 1.0,
                "coverage_annotated": 1.0,
                "incorrect": 3.0,
                "precision": 0.0,
                "total_records": 1,
            },
        ),
        (
            CreateLabelingRule(query="ejemplo", labels=["A", "o.k."]),
            {
                "annotated_records": 1,
                "correct": 1.0,
                "coverage": 1.0,
                "coverage_annotated": 1.0,
                "incorrect": 1.0,
                "precision": 0.5,
                "total_records": 1,
            },
        ),
        (
            CreateLabelingRule(query="/eje.*o/", labels=["A", "o.k."]),
            {
                "annotated_records": 1,
                "correct": 1.0,
                "coverage": 1.0,
                "coverage_annotated": 1.0,
                "incorrect": 1.0,
                "precision": 0.5,
                "total_records": 1,
            },
        ),
    ],
)
@pytest.mark.asyncio
async def test_rule_metrics_with_missing_label_for_stored_rule(
    async_client: "AsyncClient", argilla_user: User, rule, expected_metrics
):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_rule_metrics_with_missing_label_for_stored_rule"
    await log_some_records(async_client, dataset, workspace_name=argilla_user.username, annotation="o.k.")
    response = await async_client.post(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules", json=rule.dict(), params=workspace_query_params
    )
    assert response.status_code == 200

    response = await async_client.get(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules/{rule.query}/metrics", params=workspace_query_params
    )
    assert response.status_code == 200

    assert response.json() == expected_metrics


@pytest.mark.asyncio
async def test_create_rules_and_then_log(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_create_rules_and_then_log"

    await log_some_records(async_client, dataset, workspace_name=argilla_user.username, annotation="OK")
    for query in ["ejemplo", "bad query"]:
        response = await async_client.post(
            f"/api/datasets/TextClassification/{dataset}/labeling/rules",
            json=CreateLabelingRule(query=query, label="TEST", description="Description").dict(),
            params=workspace_query_params,
        )
        assert response.status_code == 200

    response = await async_client.get(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules", params=workspace_query_params
    )
    assert response.status_code == 200

    rules = list(map(LabelingRule.parse_obj, response.json()))
    assert len(rules) == 2

    await log_some_records(async_client, dataset, workspace_name=argilla_user.username, annotation="OK", delete=False)
    response = await async_client.get(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules", params=workspace_query_params
    )
    assert response.status_code == 200

    rules = list(map(LabelingRule.parse_obj, response.json()))
    assert len(rules) == 2


@pytest.mark.parametrize(
    ("rules", "expected_metrics", "annotation"),
    [
        (
            [
                CreateLabelingRule(query="ejemplo", label="TEST"),
                CreateLabelingRule(query="bad request", label="TEST"),
                CreateLabelingRule(query="other", labels=["A", "B"]),
            ],
            {
                "annotated_records": 1,
                "coverage": 1.0,
                "coverage_annotated": 1.0,
                "total_records": 1,
            },
            "OK",
        ),
        (
            [
                CreateLabelingRule(query="/eje.?plo/", label="TEST"),
                CreateLabelingRule(query="bad request", label="TEST"),
                CreateLabelingRule(query="other", labels=["A", "B"]),
            ],
            {
                "annotated_records": 0,
                "coverage": 1.0,
                "total_records": 1,
            },
            None,
        ),
        (
            [
                CreateLabelingRule(query="ejemplo", label="TEST"),
                CreateLabelingRule(query="bad request", label="bad.label"),
                CreateLabelingRule(query="other", labels=["A", "B", "good.label"]),
            ],
            {
                "annotated_records": 1,
                "coverage": 1.0,
                "coverage_annotated": 1.0,
                "total_records": 1,
            },
            "good.label",
        ),
    ],
)
@pytest.mark.asyncio
async def test_dataset_rules_metrics(
    async_client: "AsyncClient", argilla_user: User, rules, expected_metrics, annotation
):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_dataset_rules_metrics"
    await log_some_records(async_client, dataset, workspace_name=argilla_user.username, annotation=annotation)

    for rule in rules:
        response = await async_client.post(
            f"/api/datasets/TextClassification/{dataset}/labeling/rules",
            json=rule.dict(),
            params=workspace_query_params,
        )
        assert response.status_code == 200

    response = await async_client.get(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules/metrics", params=workspace_query_params
    )
    assert response.status_code == 200, response.json()

    assert response.json() == expected_metrics


@pytest.mark.asyncio
async def test_rule_metric(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_rule_metric"
    await log_some_records(async_client, dataset, workspace_name=argilla_user.username, annotation="OK")

    response = await async_client.get(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules/ejemplo/metrics?label=TEST",
        params=workspace_query_params,
    )
    assert response.status_code == 200

    metrics = LabelingRuleMetricsSummary.parse_obj(response.json())
    assert metrics.total_records == 1
    assert metrics.coverage == 1
    assert metrics.coverage_annotated == 1
    assert metrics.correct == 0
    assert metrics.incorrect == 1
    assert metrics.precision == 0

    response = await async_client.get(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules/ejemplo/metrics?label=OK",
        params=workspace_query_params,
    )
    assert response.status_code == 200

    metrics = LabelingRuleMetricsSummary.parse_obj(response.json())
    assert metrics.correct == 1
    assert metrics.incorrect == 0
    assert metrics.precision == 1

    response = await async_client.get(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules/ejemplo/metrics", params=workspace_query_params
    )
    assert response.status_code == 200

    metrics = LabelingRuleMetricsSummary.parse_obj(response.json())
    assert metrics.correct == 0
    assert metrics.incorrect == 0
    assert metrics.precision is None
    assert metrics.coverage_annotated == 1

    response = await async_client.get(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules/badd/metrics?label=OK",
        params=workspace_query_params,
    )
    assert response.status_code == 200

    metrics = LabelingRuleMetricsSummary.parse_obj(response.json())
    assert metrics.total_records == 1
    assert metrics.coverage == 0
    assert metrics.coverage_annotated == 0
    assert metrics.correct == 0
    assert metrics.incorrect == 0
    assert metrics.precision is None


@pytest.mark.asyncio
async def test_rule_metric_with_multiple_labels(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_rule_metric"
    await log_some_records(async_client, dataset, workspace_name=argilla_user.username, annotation="OK")

    response = await async_client.get(
        f"/api/datasets/TextClassification/{dataset}/labeling/rules/ejemplo/metrics?label=A&label=OK",
        params=workspace_query_params,
    )
    assert response.status_code == 200

    assert response.json() == {
        "annotated_records": 1,
        "correct": 1.0,
        "coverage": 1.0,
        "coverage_annotated": 1.0,
        "incorrect": 1.0,
        "precision": 0.5,
        "total_records": 1,
    }


@pytest.mark.asyncio
async def test_search_records_with_uncovered_rules(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_search_records_with_uncovered_rules"
    await log_some_records(async_client, dataset, workspace_name=argilla_user.username, annotation="OK")

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:search", params=workspace_query_params
    )
    assert response.status_code == 200

    assert len(response.json()["records"]) == 1

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:search",
        json={"query": {"uncovered_by_rules": ["texto"]}},
        params=workspace_query_params,
    )
    assert response.status_code == 200
    assert len(response.json()["records"]) == 0



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio
from typing import TYPE_CHECKING, Any, Dict, Optional

import pytest
import pytest_asyncio
from argilla_server.models.base import DatabaseModel
from sqlalchemy import JSON, inspect, select
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import Mapped, mapped_column

from tests.pydantic_v1 import BaseModel

if TYPE_CHECKING:
    from sqlalchemy.ext.asyncio import AsyncSession


class Model(DatabaseModel):
    __tablename__ = "model"

    id: Mapped[int] = mapped_column(primary_key=True)
    str_col: Mapped[Optional[str]] = mapped_column()
    int_col: Mapped[Optional[int]] = mapped_column()
    external_id: Mapped[Optional[str]] = mapped_column(unique=True)
    dict_col: Mapped[Optional[Dict[str, Any]]] = mapped_column(MutableDict.as_mutable(JSON))

    __upsertable_columns__ = {"str_col", "int_col"}


class ModelCreateSchema(BaseModel):
    str_col: str
    int_col: int
    external_id: str


@pytest_asyncio.fixture(autouse=True)
async def configure_model_table(connection: "AsyncConnection"):
    await connection.run_sync(Model.__table__.create)

    yield Model

    try:
        await connection.run_sync(Model.__table__.drop)
    except Exception:
        pass


@pytest.mark.asyncio
class TestDatabaseModel:
    async def test_database_model_create(self, db: "AsyncSession"):
        model = await Model.create(db, str_col="unit-test", int_col=1, autocommit=True)
        assert isinstance(model.id, int)
        assert model.id is not None
        assert model.str_col == "unit-test"
        assert model.int_col == 1
        assert model.inserted_at == model.updated_at

    async def test_database_model_create_without_autocommit(self, db: "AsyncSession"):
        model = await Model.create(db, autocommit=False)
        assert inspect(model).pending

    async def test_database_model_read_by(self, db: "AsyncSession"):
        await Model.create(db, str_col="unit-test", int_col=1, autocommit=True)
        model = await Model.read_by(db, str_col="unit-test")
        assert model.id is not None
        assert model.str_col == "unit-test"
        assert model.int_col == 1

    async def test_database_model_read(self, db: "AsyncSession"):
        model = await Model.create(db, str_col="unit-test", int_col=1, autocommit=True)
        model = await Model.read(db, model.id)
        assert model.id is not None
        assert model.str_col == "unit-test"
        assert model.int_col == 1

    async def test_database_model_update(self, db: "AsyncSession"):
        model = await Model.create(
            db, str_col="unit-test", int_col=1, dict_col={"a": 1, "b": 2, "c": 3}, autocommit=True
        )

        model = await model.update(db, str_col="unit-test-2", int_col=2, dict_col={"a": 10}, autocommit=True)
        assert model.str_col == "unit-test-2"
        assert model.int_col == 2
        assert model.dict_col == {"a": 10, "b": 2, "c": 3}

        model = await model.update(
            db, str_col="unit-test-2", int_col=2, dict_col={"a": 10}, replace_dict=True, autocommit=True
        )
        assert model.str_col == "unit-test-2"
        assert model.int_col == 2
        assert model.dict_col == {"a": 10}

    async def test_datbase_model_update_many(self, db: "AsyncSession"):
        model_1 = await Model.create(
            db, str_col="unit-test-1", int_col=1, dict_col={"a": 1, "b": 2, "c": 3}, autocommit=True
        )
        model_2 = await Model.create(
            db, str_col="unit-test-2", int_col=2, dict_col={"a": 4, "b": 5, "c": 6}, autocommit=True
        )
        model_3 = await Model.create(
            db, str_col="unit-test-3", int_col=3, dict_col={"a": 7, "b": 8, "c": 9}, autocommit=True
        )

        await Model.update_many(
            db,
            objects=[
                {"id": model_1.id, "int_col": 4},
                {"id": model_2.id, "int_col": 5},
                {"id": model_3.id, "int_col": 6},
            ],
        )

        assert model_1.int_col == 4
        assert model_2.int_col == 5
        assert model_3.int_col == 6

    async def test_database_model_upsert_many(self, db: "AsyncSession"):
        models = []
        schemas = []
        # These ones will be updated
        for i in range(5):
            models.append(await Model.create(db, str_col=f"unit-test-{i}", int_col=i, external_id=f"external-id-{i}"))
            schemas.append(
                ModelCreateSchema(str_col=f"unit-test-{i}-updated", int_col=i * 10, external_id=f"external-id-{i}")
            )
        # This one has to be inserted
        schemas.append(
            ModelCreateSchema(str_col="unit-test-inserted", int_col=99999, external_id="external-id-inserted")
        )
        models = await Model.upsert_many(db, schemas, constraints=[Model.external_id], autocommit=True)
        for i, model in enumerate(models[:5]):
            assert model.str_col == f"unit-test-{i}-updated"
            assert model.int_col == i * 10
        assert models[-1].str_col == "unit-test-inserted"
        assert models[-1].int_col == 99999

    async def test_database_model_upsert_many_without_objects(self, db: "AsyncSession"):
        with pytest.raises(ValueError, match="Cannot upsert empty list of objects"):
            await Model.upsert_many(db, [], constraints=[Model.external_id], autocommit=True)

    async def test_database_model_upsert(self, db: "AsyncSession"):
        model = await Model.create(db, str_col="unit-test", int_col=1, external_id="12345", autocommit=True)
        await asyncio.sleep(1)
        model = await Model.upsert(
            db,
            ModelCreateSchema(str_col="unit-test-updated", int_col=2, external_id=model.external_id),
            constraints=[Model.external_id],
            autocommit=True,
        )
        assert model.str_col == "unit-test-updated"
        assert model.int_col == 2

    async def test_database_model_upsert_updated_at(self, db: "AsyncSession"):
        model = await Model.create(db, str_col="unit-test", int_col=1, external_id="12345", autocommit=True)
        updated_at = model.updated_at
        await asyncio.sleep(1)

        model = await Model.upsert(
            db,
            ModelCreateSchema(str_col="unit-test-updated", int_col=2, external_id=model.external_id),
            constraints=[Model.external_id],
            autocommit=True,
        )

        assert model.updated_at > updated_at

    async def test_database_model_delete(self, db: "AsyncSession"):
        model = await Model.create(db, str_col="unit-test", int_col=1, autocommit=True)
        model = await model.delete(db)
        assert (await db.execute(select(Model).filter_by(id=model.id))).scalar_one_or_none() is None

    async def test_database_model_delete_many(self, db: "AsyncSession"):
        for i in range(5):
            await Model.create(db, str_col=f"unit-test-{i}", int_col=i, external_id=f"external-id-{i}")
        for i in range(1, 6):
            await Model.create(db, str_col=f"unit-test-{i}", int_col=i * 10, external_id=f"external-id-{i * 10}")

        removed = await Model.delete_many(db, params=[Model.int_col < 10], autocommit=True)

        assert len(removed) == 5



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Dict, Set, Union

from argilla_server.pydantic_v1 import BaseModel, root_validator


class UpdateSchema(BaseModel):
    """Base schema for update endpoints. `__non_nullable_fields__` is a set of fields that cannot be set to `None`
    explicitly. The list of fields is validated in `validate_non_nullable_fields` root validator, which will raise a
    `ValueError` if any of the fields in the set was set to `None` explicitly.
    """

    __non_nullable_fields__: Union[Set[str], None] = None

    @root_validator(pre=True)
    def validate_non_nullable_fields(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if cls.__non_nullable_fields__ is None:
            return values

        invalid_keys = []
        for key in cls.__non_nullable_fields__:
            if key in values and values[key] is None:
                invalid_keys.append(key)

        if invalid_keys:
            raise ValueError(f"The following keys must have non-null values: {', '.join(invalid_keys)}")

        return values



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import os

import pytest
from argilla_server.commons.models import TaskType
from argilla_server.daos.backend import GenericElasticEngineBackend
from argilla_server.daos.backend.generic_elastic import dataset_records_index


@pytest.mark.skipif("GITHUB_RUN_ID" in os.environ, reason="This test fails often in GitHub actions")
def test_copy_index_as_alias(es: GenericElasticEngineBackend, opensearch):
    source_id = "source_id"
    source_id_alias = f"{source_id}_alias"
    target_id = "target_id"

    es.delete(source_id)
    es.delete(target_id)

    opensearch.indices.refresh()

    es.create_dataset(id=source_id, task=TaskType.text_classification)
    es.client.create_index_alias(
        index=dataset_records_index(source_id),
        alias=dataset_records_index(f"{source_id}_alias"),
    )

    assert es.get_schema(source_id) == es.get_schema(source_id_alias)

    es.copy(id_from=source_id_alias, id_to=target_id)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from uuid import uuid4

import argilla_server.errors.future as errors
import pytest
from argilla_server.contexts.search import SearchRecordsQueryValidator
from argilla_server.schemas.v1.records import SearchRecordsQuery
from sqlalchemy.ext.asyncio import AsyncSession

from tests.factories import (
    DatasetFactory,
    FloatMetadataPropertyFactory,
    LabelSelectionQuestionFactory,
    TextQuestionFactory,
)


@pytest.mark.asyncio
class TestSearchRecordsQueryValidator:
    async def test_validate(self, db: AsyncSession):
        dataset = await DatasetFactory.create()
        text_question = await TextQuestionFactory.create(dataset=dataset)
        label_selection_question = await LabelSelectionQuestionFactory.create(dataset=dataset)
        metadata_property = await FloatMetadataPropertyFactory.create(dataset=dataset)

        query = SearchRecordsQuery.parse_obj(
            {
                "query": {
                    "text": {"q": "query"},
                },
                "filters": {
                    "and": [
                        {
                            "type": "terms",
                            "scope": {"entity": "response", "question": text_question.name, "property": "status"},
                            "values": ["submitted", "draft"],
                        },
                        {
                            "type": "terms",
                            "scope": {"entity": "suggestion", "question": text_question.name, "property": "agent"},
                            "values": ["GPT-3.5", "GPT-4"],
                        },
                        {
                            "type": "terms",
                            "scope": {"entity": "suggestion", "question": label_selection_question.name},
                            "values": ["politics", "news"],
                        },
                        {
                            "type": "range",
                            "scope": {"entity": "suggestion", "question": text_question.name, "property": "score"},
                            "ge": 0.8,
                            "le": 1.0,
                        },
                        {
                            "type": "range",
                            "scope": {"entity": "metadata", "metadata_property": metadata_property.name},
                            "ge": 0.5,
                            "le": 0.8,
                        },
                    ]
                },
                "sort": [
                    {"scope": {"entity": "record", "property": "inserted_at"}, "order": "desc"},
                    {"scope": {"entity": "record", "property": "updated_at"}, "order": "desc"},
                    {
                        "scope": {"entity": "suggestion", "question": text_question.name, "property": "score"},
                        "order": "desc",
                    },
                    {"scope": {"entity": "metadata", "metadata_property": metadata_property.name}, "order": "asc"},
                ],
            }
        )

        await SearchRecordsQueryValidator(db, query, dataset.id).validate()

    async def test_validate_response_filter_scope_in_filters_without_question(self, db: AsyncSession):
        query = SearchRecordsQuery.parse_obj(
            {
                "query": {
                    "text": {"q": "query"},
                },
                "filters": {"and": [{"type": "terms", "scope": {"entity": "response"}, "values": ["value"]}]},
            }
        )

        await SearchRecordsQueryValidator(db, query, uuid4()).validate()

    async def test_validate_response_filter_scope_in_filters_with_non_existent_question(self, db: AsyncSession):
        dataset = await DatasetFactory.create()

        query = SearchRecordsQuery.parse_obj(
            {
                "query": {
                    "text": {"q": "query"},
                },
                "filters": {
                    "and": [
                        {
                            "type": "terms",
                            "scope": {"entity": "response", "question": "non-existent"},
                            "values": ["value"],
                        }
                    ]
                },
            }
        )

        with pytest.raises(errors.NotFoundError) as not_found_error:
            await SearchRecordsQueryValidator(db, query, dataset.id).validate()

        assert (
            str(not_found_error.value)
            == f"Question with name `non-existent` not found for dataset with id `{dataset.id}`"
        )

    async def test_validate_suggestion_filter_scope_in_filters_with_non_existent_question(self, db: AsyncSession):
        dataset = await DatasetFactory.create()

        query = SearchRecordsQuery.parse_obj(
            {
                "query": {
                    "text": {"q": "query"},
                },
                "filters": {
                    "and": [
                        {
                            "type": "terms",
                            "scope": {"entity": "suggestion", "question": "non-existent"},
                            "values": ["value"],
                        }
                    ]
                },
            }
        )

        with pytest.raises(errors.NotFoundError) as not_found_error:
            await SearchRecordsQueryValidator(db, query, dataset.id).validate()

        assert (
            str(not_found_error.value)
            == f"Question with name `non-existent` not found for dataset with id `{dataset.id}`"
        )

    async def test_validate_metadata_filter_scope_in_filters_with_non_existent_metadata_property(
        self, db: AsyncSession
    ):
        dataset = await DatasetFactory.create()

        query = SearchRecordsQuery.parse_obj(
            {
                "query": {
                    "text": {"q": "query"},
                },
                "filters": {
                    "and": [
                        {
                            "type": "terms",
                            "scope": {"entity": "metadata", "metadata_property": "non-existent"},
                            "values": ["value"],
                        }
                    ]
                },
            }
        )

        with pytest.raises(errors.NotFoundError) as not_found_error:
            await SearchRecordsQueryValidator(db, query, dataset.id).validate()

        assert (
            str(not_found_error.value)
            == f"Metadata property with name `non-existent` not found for dataset with id `{dataset.id}`"
        )

    async def test_validate_response_filter_scope_in_sort_without_question(self, db: AsyncSession):
        query = SearchRecordsQuery.parse_obj(
            {
                "query": {
                    "text": {"q": "query"},
                },
                "sort": [{"scope": {"entity": "response"}, "order": "asc"}],
            }
        )

        await SearchRecordsQueryValidator(db, query, uuid4()).validate()

    async def test_validate_response_filter_scope_in_sort_with_non_existent_question(self, db: AsyncSession):
        dataset = await DatasetFactory.create()

        query = SearchRecordsQuery.parse_obj(
            {
                "query": {
                    "text": {"q": "query"},
                },
                "sort": [{"scope": {"entity": "response", "question": "non-existent"}, "order": "asc"}],
            }
        )

        with pytest.raises(errors.NotFoundError) as not_found_error:
            await SearchRecordsQueryValidator(db, query, dataset.id).validate()

        assert (
            str(not_found_error.value)
            == f"Question with name `non-existent` not found for dataset with id `{dataset.id}`"
        )

    async def test_validate_suggestion_filter_scope_in_sort_with_non_existent_question(self, db: AsyncSession):
        dataset = await DatasetFactory.create()

        query = SearchRecordsQuery.parse_obj(
            {
                "query": {
                    "text": {"q": "query"},
                },
                "sort": [{"scope": {"entity": "suggestion", "question": "non-existent"}, "order": "asc"}],
            }
        )

        with pytest.raises(errors.NotFoundError) as not_found_error:
            await SearchRecordsQueryValidator(db, query, dataset.id).validate()

        assert (
            str(not_found_error.value)
            == f"Question with name `non-existent` not found for dataset with id `{dataset.id}`"
        )

    async def test_validate_metadata_filter_scope_in_sort_with_non_existent_metadata_property(self, db: AsyncSession):
        dataset = await DatasetFactory.create()

        query = SearchRecordsQuery.parse_obj(
            {
                "query": {
                    "text": {"q": "query"},
                },
                "sort": [{"scope": {"entity": "metadata", "metadata_property": "non-existent"}, "order": "asc"}],
            }
        )

        with pytest.raises(errors.NotFoundError) as not_found_error:
            await SearchRecordsQueryValidator(db, query, dataset.id).validate()

        assert (
            str(not_found_error.value)
            == f"Metadata property with name `non-existent` not found for dataset with id `{dataset.id}`"
        )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import TYPE_CHECKING, Optional

import pytest
from argilla_server.commons.models import TaskType
from argilla_server.constants import API_KEY_HEADER_NAME
from argilla_server.models import User

if TYPE_CHECKING:
    from httpx import AsyncClient


async def create_dataset(client: "AsyncClient", name: str, workspace_name: str):
    response = await client.post(
        "/api/datasets", json={"name": name, "workspace": workspace_name, f"task": TaskType.token_classification}
    )
    assert response.status_code == 200


async def delete_dataset(client: "AsyncClient", name: str, workspace_name: str):
    response = await client.delete(f"/api/datasets/{name}", params={"workspace": workspace_name})
    assert response.status_code == 200


async def create_settings(async_client: "AsyncClient", name: str, workspace_name: str, labels: Optional[list] = None):
    response = await async_client.put(
        f"/api/datasets/{TaskType.token_classification}/{name}/settings",
        json={"label_schema": {"labels": labels or ["Label1", "Label2"]}},
        params={"workspace": workspace_name},
    )
    return response


async def fetch_settings(async_client: "AsyncClient", name: str, workspace_name: str):
    return await async_client.get(
        f"/api/datasets/{TaskType.token_classification}/{name}/settings", params={"workspace": workspace_name}
    )


async def delete_dataset_settings(async_client: "AsyncClient", name: str, workspace_name: str):
    response = await async_client.delete(
        f"/api/datasets/{TaskType.token_classification}/{name}/settings", params={"workspace": workspace_name}
    )
    assert response.status_code == 200


async def log_some_data(async_client: "AsyncClient", name: str, workspace_name: str):
    response = await async_client.post(
        f"/api/datasets/{name}/TokenClassification:bulk",
        json={
            "records": [
                {
                    "tokens": "This is a text".split(" "),
                    "raw_text": "This is a text",
                    "prediction": {
                        "agent": "test",
                        "entities": [{"start": 0, "end": 4, "label": "BAD"}],
                    },
                    "annotation": {
                        "agent": "test",
                        "entities": [{"start": 0, "end": 4, "label": "BAD"}],
                    },
                }
            ]
        },
        params={"workspace": workspace_name},
    )
    return response


@pytest.mark.parametrize("labels", [["Label1", "Label2"], ["1", "2", "3"], [1, 2, 3, 4]])
@pytest.mark.asyncio
async def test_create_dataset_settings(async_client: "AsyncClient", argilla_user: User, labels: list):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_name = argilla_user.username

    name = "test_create_dataset_settings"
    await delete_dataset(async_client, name, workspace_name=workspace_name)
    await create_dataset(async_client, name, workspace_name=workspace_name)

    response = await create_settings(async_client, name, workspace_name=workspace_name, labels=labels)
    assert response.status_code == 200

    created = response.json()
    response = await fetch_settings(async_client, name, workspace_name=workspace_name)
    assert response.json() == created


@pytest.mark.asyncio
async def test_get_dataset_settings_not_found(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_name = argilla_user.username

    name = "test_get_dataset_settings_not_found"
    await delete_dataset(async_client, name, workspace_name=workspace_name)
    await create_dataset(async_client, name, workspace_name=workspace_name)

    response = await fetch_settings(async_client, name, workspace_name=workspace_name)
    assert response.status_code == 404


@pytest.mark.asyncio
async def test_delete_settings(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_name = argilla_user.username

    name = "test_delete_settings"
    await delete_dataset(async_client, name, workspace_name=workspace_name)
    await create_dataset(async_client, name, workspace_name=workspace_name)

    response = await create_settings(async_client, name, workspace_name)
    assert response.status_code == 200

    await delete_dataset_settings(async_client, name, workspace_name)

    response = await fetch_settings(async_client, name, workspace_name)
    assert response.status_code == 404


@pytest.mark.asyncio
async def test_validate_settings_when_logging_data(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_name = argilla_user.username

    name = "test_validate_settings_when_logging_data"
    await delete_dataset(async_client, name, workspace_name=workspace_name)
    await create_dataset(async_client, name, workspace_name=workspace_name)

    response = await create_settings(async_client, name, workspace_name)
    assert response.status_code == 200

    response = await log_some_data(async_client, name, workspace_name)
    assert response.status_code == 400

    assert response.json() == {
        "detail": {
            "code": "argilla.api.errors::BadRequestError",
            "params": {
                "message": "Provided records contain the BAD label, "
                "that is not included in the labels schema.\n"
                "Please, annotate your records using labels "
                "defined in the labels schema."
            },
        }
    }


@pytest.mark.asyncio
async def test_validate_settings_after_logging(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_name = argilla_user.username

    name = "test_validate_settings_after_logging"
    await delete_dataset(async_client, name, workspace_name=workspace_name)
    await create_dataset(async_client, name, workspace_name=workspace_name)

    response = await log_some_data(async_client, name, workspace_name)
    assert response.status_code == 200

    response = await create_settings(async_client, name, workspace_name)
    assert response.status_code == 400

    assert response.json() == {
        "detail": {
            "code": "argilla.api.errors::BadRequestError",
            "params": {
                "message": "The label BAD was found in the dataset but "
                "not in provided labels schema. \n"
                "Please, provide a valid labels schema "
                "according to stored records in the "
                "dataset"
            },
        }
    }



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import contextlib
import uuid
from typing import TYPE_CHECKING, Dict, Generator

import pytest
import pytest_asyncio
from argilla_server import telemetry
from argilla_server.constants import API_KEY_HEADER_NAME, DEFAULT_API_KEY
from argilla_server.daos.backend import GenericElasticEngineBackend
from argilla_server.daos.datasets import DatasetsDAO
from argilla_server.daos.records import DatasetRecordsDAO
from argilla_server.database import get_async_db
from argilla_server.models import User, UserRole, Workspace
from argilla_server.search_engine import SearchEngine, get_search_engine
from argilla_server.settings import settings
from argilla_server.telemetry import TelemetryClient
from httpx import AsyncClient
from opensearchpy import OpenSearch

from tests.database import TestSession
from tests.factories import AnnotatorFactory, OwnerFactory, UserFactory

if TYPE_CHECKING:
    from unittest.mock import MagicMock

    from pytest_mock import MockerFixture


@pytest.fixture(scope="session")
def elasticsearch_config():
    return {"hosts": settings.elasticsearch}


@pytest.fixture(scope="session", autouse=True)
def opensearch(elasticsearch_config: dict) -> Generator[OpenSearch, None, None]:
    client = OpenSearch(**elasticsearch_config)
    yield client

    for index_info in client.cat.indices(index="ar.*,rg.*", format="json"):
        client.indices.delete(index=index_info["index"])


@pytest.fixture(scope="session")
def es():
    return GenericElasticEngineBackend.get_instance()


@pytest.fixture(scope="function")
def mock_search_engine(mocker) -> Generator["SearchEngine", None, None]:
    return mocker.AsyncMock(SearchEngine)


@pytest_asyncio.fixture(scope="function")
async def owner() -> User:
    return await OwnerFactory.create(first_name="Owner", username="owner", api_key="owner.apikey")


@pytest_asyncio.fixture(scope="function")
async def annotator() -> User:
    return await AnnotatorFactory.create(first_name="Annotator", username="annotator", api_key="annotator.apikey")


@pytest.fixture(scope="function")
def owner_auth_header(owner: User) -> Dict[str, str]:
    return {API_KEY_HEADER_NAME: owner.api_key}


@pytest_asyncio.fixture(scope="function")
async def async_client(
    request, mock_search_engine: SearchEngine, mocker: "MockerFixture"
) -> Generator["AsyncClient", None, None]:
    from argilla_server import app

    async def override_get_async_db():
        session = TestSession()
        yield session

    async def override_get_search_engine():
        yield mock_search_engine

    mocker.patch("argilla_server._app._get_db_wrapper", wraps=contextlib.asynccontextmanager(override_get_async_db))

    app.dependency_overrides[get_async_db] = override_get_async_db
    app.dependency_overrides[get_search_engine] = override_get_search_engine

    async with AsyncClient(app=app, base_url="http://testserver") as async_client:
        yield async_client

    app.dependency_overrides.clear()


@pytest.fixture(autouse=True)
def test_telemetry(mocker: "MockerFixture") -> "MagicMock":
    mock_telemetry = mocker.Mock(TelemetryClient)
    mock_telemetry.server_id = uuid.uuid4()

    telemetry._CLIENT = mock_telemetry
    return telemetry._CLIENT


@pytest.fixture(scope="session")
def records_dao(es: GenericElasticEngineBackend):
    return DatasetRecordsDAO.get_instance(es)


@pytest.fixture(scope="session")
def datasets_dao(records_dao: DatasetRecordsDAO, es: GenericElasticEngineBackend):
    return DatasetsDAO.get_instance(es=es, records_dao=records_dao)


@pytest_asyncio.fixture(scope="function")
async def argilla_user() -> Generator[User, None, None]:
    user = await UserFactory.create(
        first_name="Argilla",
        username="argilla",
        role=UserRole.admin,  # Force to use an admin user
        password_hash="$2y$05$eaw.j2Kaw8s8vpscVIZMfuqSIX3OLmxA21WjtWicDdn0losQ91Hw.",
        api_key=DEFAULT_API_KEY,
        workspaces=[Workspace(name="argilla")],
    )
    yield user
